"""Unit tests for :class:`esmf_regrid.schemes.ESMFAreaWeighted`."""

import pytest

from esmf_regrid.schemes import ESMFAreaWeighted
from esmf_regrid.tests.unit.schemes.__init__ import (
    _test_cube_regrid,
    _test_dtype_handling,
    _test_invalid_mdtol,
    _test_mask_from_init,
    _test_mask_from_regridder,
    _test_non_degree_crs,
)


@pytest.mark.parametrize(
    "src_type,tgt_type",
    [
        ("grid", "grid"),
        ("grid", "mesh"),
        ("grid", "just_mesh"),
        ("mesh", "grid"),
        ("mesh", "mesh"),
    ],
)
def test_cube_regrid(src_type, tgt_type):
    """
    Test that ESMFAreaWeighted can be passed to a cubes regrid method.

    Checks that regridding occurs and that mdtol is used correctly.
    """
    _test_cube_regrid(ESMFAreaWeighted, src_type, tgt_type)


def test_invalid_mdtol():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFAreaWeighted`.

    Checks that an error is raised when mdtol is out of range.
    """
    _test_invalid_mdtol(ESMFAreaWeighted)


@pytest.mark.parametrize("mask_keyword", ["use_src_mask", "use_tgt_mask"])
def test_mask_from_init(mask_keyword):
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFAreaWeighted`.

    Checks that use_src_mask and use_tgt_mask are passed down correctly.
    """
    _test_mask_from_init(ESMFAreaWeighted, mask_keyword)


@pytest.mark.parametrize("mask_keyword", ["use_src_mask", "use_tgt_mask"])
def test_mask_from_regridder(mask_keyword):
    """
    Test regridder method of :class:`esmf_regrid.schemes.ESMFAreaWeighted`.

    Checks that use_src_mask and use_tgt_mask are passed down correctly.
    """
    _test_mask_from_regridder(ESMFAreaWeighted, mask_keyword)


def test_invalid_tgt_location():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFAreaWeighted`.

    Checks that initialisation fails when tgt_location is not "face".
    """
    match = "For area weighted regridding, target location must be 'face'."
    with pytest.raises(ValueError, match=match):
        _ = ESMFAreaWeighted(tgt_location="node")


def test_non_degree_crs():
    """Test for coordinates with non-degree units."""
    _test_non_degree_crs(ESMFAreaWeighted)


def test_dtype_handling(src_tgt_types, in_dtype):
    """Test regridding scheme handles dtype as expected."""
    src_type, tgt_type = src_tgt_types
    _test_dtype_handling(ESMFAreaWeighted, src_type, tgt_type, in_dtype)



try:
    import esmpy
except ImportError as exc:
    # Prior to v8.4.0, `esmpy`` could be imported as `ESMF`.
    try:
        import ESMF as esmpy  # noqa: N811
    except ImportError:
        raise exc

# constants needs to be above schemes, as it is used within
from .constants import Constants, check_method, check_norm
from .schemes import *


__version__ = "0.10.0"



"""Unit tests for :class:`esmf_regrid.schemes.ESMFNearest`."""

import numpy as np
from numpy import ma
import pytest

from esmf_regrid.schemes import ESMFNearest
from esmf_regrid.tests.unit.schemes.__init__ import (
    _test_dtype_handling,
    _test_mask_from_init,
    _test_mask_from_regridder,
    _test_non_degree_crs,
)
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import _grid_cube
from esmf_regrid.tests.unit.schemes.test__mesh_to_MeshInfo import (
    _gridlike_mesh,
    _gridlike_mesh_cube,
)


@pytest.mark.parametrize(
    "src_type,tgt_type",
    [
        ("grid", "grid"),
        ("grid", "mesh"),
        ("grid", "just_mesh"),
        ("mesh", "grid"),
        ("mesh", "mesh"),
    ],
)
def test_cube_regrid(src_type, tgt_type):
    """
    Test that ESMFNearest can be passed to a cubes regrid method.

    Checks that regridding occurs.
    """
    if tgt_type == "just_mesh":
        scheme_default = ESMFNearest(tgt_location="face")
    else:
        scheme_default = ESMFNearest()

    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    if src_type == "grid":
        src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
        src_data = np.zeros([n_lats_src, n_lons_src])
        src_mask = np.zeros([n_lats_src, n_lons_src])
        src_mask[0, 0] = 1
    else:
        src = _gridlike_mesh_cube(n_lons_src, n_lats_src)
        src_data = np.zeros([n_lats_src * n_lons_src])
        src_mask = np.zeros([n_lats_src * n_lons_src])
        src_mask[0] = 1
    if tgt_type == "grid":
        tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
        expected_data_default = np.zeros([n_lats_tgt, n_lons_tgt])
        expected_mask = np.zeros([n_lats_tgt, n_lons_tgt])
        expected_mask[0, 0] = 1
    elif tgt_type == "mesh":
        tgt = _gridlike_mesh_cube(n_lons_tgt, n_lats_tgt)
        expected_data_default = np.zeros([n_lats_tgt * n_lons_tgt])
        expected_mask = np.zeros([n_lats_tgt * n_lons_tgt])
        expected_mask[0] = 1
    elif tgt_type == "just_mesh":
        tgt = _gridlike_mesh(n_lons_tgt, n_lats_tgt)
        expected_data_default = np.zeros([n_lats_tgt * n_lons_tgt])
        expected_mask = np.zeros([n_lats_tgt * n_lons_tgt])
        expected_mask[0] = 1
    src_data = ma.array(src_data, mask=src_mask)
    src.data = src_data

    result_default = src.regrid(tgt, scheme_default)

    if tgt_type == "just_mesh":
        expected_cube_default = _gridlike_mesh_cube(n_lons_tgt, n_lats_tgt)
    else:
        expected_cube_default = tgt.copy()
    expected_cube_default.data = expected_data_default

    assert expected_cube_default == result_default


@pytest.mark.parametrize("mask_keyword", ["use_src_mask", "use_tgt_mask"])
def test_mask_from_init(mask_keyword):
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFNearest`.

    Checks that use_src_mask and use_tgt_mask are passed down correctly.
    """
    _test_mask_from_init(ESMFNearest, mask_keyword)


@pytest.mark.parametrize("mask_keyword", ["use_src_mask", "use_tgt_mask"])
def test_mask_from_regridder(mask_keyword):
    """
    Test regridder method of :class:`esmf_regrid.schemes.ESMFNearest`.

    Checks that use_src_mask and use_tgt_mask are passed down correctly.
    """
    _test_mask_from_regridder(ESMFNearest, mask_keyword)


def test_non_degree_crs():
    """Test for coordinates with non-degree units."""
    _test_non_degree_crs(ESMFNearest)


def test_dtype_handling(src_tgt_types, in_dtype):
    """Test regridding scheme handles dtype as expected."""
    src_type, tgt_type = src_tgt_types
    _test_dtype_handling(ESMFNearest, src_type, tgt_type, in_dtype)



"""Unit tests for `esmf_regrid.schemes`."""

import dask.array as da
from iris.coord_systems import OSGB
import numpy as np
from numpy import ma
import pytest

from esmf_regrid.schemes import ESMFAreaWeighted, ESMFBilinear, ESMFNearest
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import _grid_cube
from esmf_regrid.tests.unit.schemes.test__mesh_to_MeshInfo import (
    _gridlike_mesh,
    _gridlike_mesh_cube,
)


def _test_cube_regrid(scheme, src_type, tgt_type):
    """
    Test that `scheme` can be passed to a cubes regrid method.

    Checks that regridding occurs and that mdtol is used correctly.
    """
    if tgt_type == "just_mesh":
        scheme_default = scheme(tgt_location="face")
        scheme_full_mdtol = scheme(mdtol=1, tgt_location="face")
    else:
        scheme_default = scheme()
        scheme_full_mdtol = scheme(mdtol=1)

    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    if src_type == "grid":
        src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
        src_data = np.zeros([n_lats_src, n_lons_src])
        src_mask = np.zeros([n_lats_src, n_lons_src])
        src_mask[0, 0] = 1
    else:
        src = _gridlike_mesh_cube(n_lons_src, n_lats_src)
        src_data = np.zeros([n_lats_src * n_lons_src])
        src_mask = np.zeros([n_lats_src * n_lons_src])
        src_mask[0] = 1
    if tgt_type == "grid":
        tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
        expected_data_default = np.zeros([n_lats_tgt, n_lons_tgt])
        expected_mask = np.zeros([n_lats_tgt, n_lons_tgt])
        expected_mask[0, 0] = 1
    elif tgt_type == "mesh":
        tgt = _gridlike_mesh_cube(n_lons_tgt, n_lats_tgt)
        expected_data_default = np.zeros([n_lats_tgt * n_lons_tgt])
        expected_mask = np.zeros([n_lats_tgt * n_lons_tgt])
        expected_mask[0] = 1
    elif tgt_type == "just_mesh":
        tgt = _gridlike_mesh(n_lons_tgt, n_lats_tgt)
        expected_data_default = np.zeros([n_lats_tgt * n_lons_tgt])
        expected_mask = np.zeros([n_lats_tgt * n_lons_tgt])
        expected_mask[0] = 1

    src_data = ma.array(src_data, mask=src_mask)
    src.data = src_data

    result_default = src.regrid(tgt, scheme_default)
    result_full = src.regrid(tgt, scheme_full_mdtol)

    expected_data_full = ma.array(expected_data_default, mask=expected_mask)

    if tgt_type == "just_mesh":
        tgt_template = _gridlike_mesh_cube(n_lons_tgt, n_lats_tgt)
    else:
        tgt_template = tgt
    expected_cube_default = tgt_template.copy()
    expected_cube_default.data = expected_data_default

    expected_cube_full = tgt_template.copy()
    expected_cube_full.data = expected_data_full

    assert expected_cube_default == result_default
    assert expected_cube_full == result_full


def _test_invalid_mdtol(scheme):
    """
    Test initialisation of the scheme.

    Checks that an error is raised when mdtol is out of range.
    """
    match = "Value for mdtol must be in range 0 - 1, got "
    with pytest.raises(ValueError, match=match):
        _ = scheme(mdtol=2)
    with pytest.raises(ValueError, match=match):
        _ = scheme(mdtol=-1)


def _test_mask_from_init(scheme, mask_keyword):
    """
    Test initialisation of scheme.

    Checks that use_src_mask and use_tgt_mask are passed down correctly.
    """
    # Create a scheme with and without masking behaviour
    kwargs = {mask_keyword: True}
    default_scheme = scheme()
    masked_scheme = scheme(**kwargs)
    assert getattr(default_scheme, mask_keyword) is False
    assert getattr(masked_scheme, mask_keyword) is True

    n_lons_src = 6
    n_lats_src = 4
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    cube = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    data = np.zeros([n_lats_src, n_lons_src])
    mask = np.zeros([n_lats_src, n_lons_src])
    mask[0, 0] = 1
    data = ma.array(data, mask=mask)
    cube.data = data

    # Create a regridder from the scheme.
    default_rg = default_scheme.regridder(cube, cube)
    masked_rg = masked_scheme.regridder(cube, cube)

    # Remove "use_" from the keyword to get the equivalent attr used on the regridder.
    regridder_attr = mask_keyword[4:]

    # Check that the mask stored on the regridder is correct.
    assert getattr(default_rg, regridder_attr) is None
    np.testing.assert_allclose(getattr(masked_rg, regridder_attr), mask)


def _test_mask_from_regridder(scheme, mask_keyword):
    """
    Test regridder method of the scheme.

    Checks that use_src_mask and use_tgt_mask are passed down correctly.
    """
    n_lons_src = 6
    n_lats_src = 4
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    cube = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    data = np.zeros([n_lats_src, n_lons_src])
    mask = np.zeros([n_lats_src, n_lons_src])
    mask[0, 0] = 1
    data = ma.array(data, mask=mask)
    cube.data = data

    mask_different = np.zeros([n_lats_src, n_lons_src])
    mask_different[1, 2] = 1

    # Create a scheme without default masking behaviour.
    default_scheme = scheme()

    # Create a regridder from the mask on the cube.
    kwargs = {mask_keyword: True}
    rg_from_cube = default_scheme.regridder(cube, cube, **kwargs)

    # Create a regridder from a user supplied mask.
    kwargs_different = {mask_keyword: mask_different}
    rg_from_different = default_scheme.regridder(cube, cube, **kwargs_different)

    # Remove "use_" from the keyword to get the equivalent attr used on the regridder.
    regridder_attr = mask_keyword[4:]

    # Check that the mask stored on the regridder is correct.
    np.testing.assert_allclose(getattr(rg_from_cube, regridder_attr), mask)
    np.testing.assert_allclose(
        getattr(rg_from_different, regridder_attr), mask_different
    )


def _test_non_degree_crs(scheme):
    """Test regridding scheme is compatible with coordinates with non-degree units."""
    coord_system = OSGB()

    # This definition comes from a small section of real user data.
    n_lons_src = 2
    n_lats_src = 3
    lon_bounds = (-197500, -192500)
    lat_bounds = (1247500, 1237500)
    tm_cube = _grid_cube(
        n_lons_src,
        n_lats_src,
        lon_bounds,
        lat_bounds,
        circular=False,
        coord_system=coord_system,
        standard_names=["projection_x_coordinate", "projection_y_coordinate"],
        units="m",
    )
    data = np.arange(n_lats_src * n_lons_src).reshape([n_lats_src, n_lons_src])
    tm_cube.data = data

    n_lons_tgt = 12
    n_lats_tgt = 14
    lon_bounds_tgt = (-13, -12.8)
    lat_bounds_tgt = (60.5, 60.7)
    cube_tgt = _grid_cube(
        n_lons_tgt, n_lats_tgt, lon_bounds_tgt, lat_bounds_tgt, circular=True
    )

    result = tm_cube.regrid(cube_tgt, scheme())

    # Set expected results, this varies depending on the scheme.
    if scheme is ESMFAreaWeighted:
        expected_sum, expected_unmasked = 50.86147272655136, 21
    elif scheme is ESMFBilinear:
        expected_sum, expected_unmasked = 35.90837983047451, 13
    elif scheme is ESMFNearest:
        expected_sum, expected_unmasked = 490, 168

    # Check that the data is as expected.
    assert np.isclose(result.data.sum(), expected_sum)

    # Check that the number of masked points is as expected.
    assert (1 - result.data.mask).sum() == expected_unmasked


def _test_dtype_handling(scheme, src_type, tgt_type, in_dtype):
    """Test regridding scheme handles dtype as expected."""
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    if in_dtype == "float32":
        dtype = np.float32
    elif in_dtype == "float64":
        dtype = np.float64

    if src_type == "grid":
        src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
        src_data = np.zeros([n_lats_src, n_lons_src], dtype=dtype)
        src.data = da.array(src_data)
    elif src_type == "mesh":
        src = _gridlike_mesh_cube(n_lons_src, n_lats_src)
        src_data = np.zeros([n_lats_src * n_lons_src], dtype=dtype)
        src.data = da.array(src_data)
    if tgt_type == "grid":
        tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    elif tgt_type == "mesh":
        tgt = _gridlike_mesh_cube(n_lons_tgt, n_lats_tgt)

    result = src.regrid(tgt, scheme())

    expected_dtype = np.float64

    assert result.has_lazy_data()

    assert result.lazy_data().dtype == expected_dtype
    assert result.data.dtype == expected_dtype



"""Common testing infrastructure."""

import pytest


@pytest.fixture(params=["float32", "float64"])
def in_dtype(request):
    """Fixture for controlling dtype."""
    return request.param


@pytest.fixture(
    params=[
        ("grid", "grid"),
        ("grid", "mesh"),
        ("mesh", "grid"),
        ("mesh", "mesh"),
    ]
)
def src_tgt_types(request):
    """Fixture for controlling type of source and target."""
    return request.param



"""Provides representations of ESMF's Spatial Discretisation Objects."""

from abc import ABC, abstractmethod

import cartopy.crs as ccrs
import numpy as np
import scipy.sparse

from . import esmpy


class SDO(ABC):
    """
    Abstract base class for handling spatial discretisation objects.

    This contains shared things for representing the three spatial discretisation
    objects supported by ESMPy, Grids, Meshes, and LocStreams.
    """

    def __init__(self, shape, index_offset, field_kwargs, mask=None):
        self._shape = shape
        self._index_offset = index_offset
        self._field_kwargs = field_kwargs
        self._mask = mask

    @abstractmethod
    def _make_esmf_sdo(self):
        pass

    def make_esmf_field(self):
        """Return an ESMF field representing the spatial discretisation object."""
        sdo = self._make_esmf_sdo()
        field = esmpy.Field(sdo, **self._field_kwargs)
        return field

    @property
    def shape(self):
        """Return shape."""
        return self._shape

    @property
    def _refined_shape(self):
        """Return shape passed to ESMF."""
        return self._shape

    @property
    def _refined_mask(self):
        """Return mask passed to ESMF."""
        return self._mask

    @property
    def dims(self):
        """Return number of dimensions."""
        return len(self._shape)

    @property
    def size(self):
        """Return the number of cells in the sdo."""
        return np.prod(self._shape)

    @property
    def _refined_size(self):
        """Return the number of cells passed to ESMF."""
        return np.prod(self._refined_shape)

    @property
    def index_offset(self):
        """Return the index offset."""
        return self._index_offset

    @property
    def mask(self):
        """Return the mask."""
        return self._mask

    def _array_to_matrix(self, array):
        """
        Reshape data to a form that is compatible with weight matrices.

        The data should be presented in the form of a matrix (i.e. 2D) in order
        to be compatible with the weight matrix.
        Weight matrices deriving from ESMF use fortran ordering when flattening
        grids to determine cell indices so we use the same order for reshaping.
        We then take the transpose so that matrix multiplication happens over
        the appropriate axes.
        """
        return array.T.reshape((self.size, -1))

    def _matrix_to_array(self, array, extra_dims):
        """
        Reshape data to restore original dimensions.

        This is the inverse operation of `_array_to_matrix`.
        """
        return array.reshape((extra_dims + self._shape)[::-1]).T


class GridInfo(SDO):
    """
    Class for handling structured grids.

    This class holds information about lat-lon type grids. That is, grids
    defined by lists of latitude and longitude values for points/bounds
    (with respect to some coordinate reference system i.e. rotated pole).
    It contains methods for translating this information into :mod:`esmpy` objects.
    In particular, there are methods for representing as a
    :class:`esmpy.api.grid.Grid` and
    as a :class:`esmpy.api.field.Field` containing that
    :class:`~esmpy.api.grid.Grid`. This esmpy :class:`~esmpy.api.field.Field`
    is designed to
    contain enough information for area weighted regridding and may be
    inappropriate for other :mod:`esmpy` regridding schemes.

    """

    def __init__(
        self,
        lons,
        lats,
        lonbounds,
        latbounds,
        crs=None,
        circular=False,
        areas=None,
        mask=None,
        center=False,
    ):
        """
        Create a :class:`GridInfo` object describing the grid.

        Parameters
        ----------
        lons : :obj:`~numpy.typing.ArrayLike`
            A 1D or 2D array or list describing the longitudes of the
            grid points.
        lats : :obj:`~numpy.typing.ArrayLike`
            A 1D or 2D array or list describing the latitudes of the
            grid points.
        lonbounds : :obj:`~numpy.typing.ArrayLike`
            A 1D or 2D array or list describing the longitude bounds of
            the grid. Should have length one greater than ``lons``.
        latbounds : :obj:`~numpy.typing.ArrayLike`
            A 1D or 2D array or list describing the latitude bounds of
            the grid. Should have length one greater than ``lats``.
        crs : :class:`cartopy.crs.CRS`, optional
            Describes how to interpret the
            above arguments. If ``None``, defaults to :class:`~cartopy.crs.Geodetic`.
        circular : bool, default=False
            Describes if the final longitude bounds should
            be considered contiguous with the first.
        areas : :obj:`~numpy.typing.ArrayLike`, optional
            Array describing the areas associated with
            each face. If ``None``, then :mod:`esmpy` will use its own
            calculated areas.
        mask: :obj:`~numpy.typing.ArrayLike`, optional
            Array describing which elements :mod:`esmpy` will ignore.
        center : bool, default=False
            Describes if the center points of the grid cells are used in regridding
            calculations.

        """
        self.lons = lons
        self.lats = lats
        londims = len(self.lons.shape)
        if len(lonbounds.shape) != londims:
            msg = (
                f"The dimensionality of longitude bounds "
                f"({len(lonbounds.shape)}) is incompatible with the "
                f"dimensionality of the longitude ({londims})."
            )
            raise ValueError(msg)
        latdims = len(self.lats.shape)
        if len(latbounds.shape) != latdims:
            msg = (
                f"The dimensionality of latitude bounds "
                f"({len(latbounds.shape)}) is incompatible with the "
                f"dimensionality of the latitude ({latdims})."
            )
            raise ValueError(msg)
        if londims != latdims:
            msg = (
                f"The dimensionality of the longitude "
                f"({londims}) is incompatible with the "
                f"dimensionality of the latitude ({latdims})."
            )
            raise ValueError(msg)
        if londims not in (1, 2):
            msg = (
                f"Expected a latitude/longitude with a dimensionality "
                f"of 1 or 2, got {londims}."
            )
            raise ValueError(msg)
        if londims == 1:
            shape = (len(lats), len(lons))
        else:
            shape = self.lons.shape

        self.lonbounds = lonbounds
        self._refined_lonbounds = lonbounds
        self.latbounds = latbounds
        self._refined_latbounds = latbounds
        if crs is None:
            self.crs = ccrs.Geodetic()
        else:
            self.crs = crs
        self.circular = circular
        self.areas = areas
        self.center = center
        super().__init__(
            shape=shape,
            index_offset=1,
            field_kwargs={"staggerloc": esmpy.StaggerLoc.CENTER},
            mask=mask,
        )

    def _as_esmf_info(self):
        shape = np.array(self._refined_shape)

        londims = len(self.lons.shape)

        if londims == 1:
            if self.circular:
                adjustedlonbounds = self._refined_lonbounds[:-1]
            else:
                adjustedlonbounds = self._refined_lonbounds
            centerlons, centerlats = np.meshgrid(self.lons, self.lats)
            cornerlons, cornerlats = np.meshgrid(
                adjustedlonbounds, self._refined_latbounds
            )
        elif londims == 2:
            if self.circular:
                slice = np.s_[:, :-1]
            else:
                slice = np.s_[:]
            centerlons = self.lons[slice]
            centerlats = self.lats[slice]
            cornerlons = self._refined_lonbounds[slice]
            cornerlats = self._refined_latbounds[slice]

        truecenters = ccrs.Geodetic().transform_points(self.crs, centerlons, centerlats)
        truecorners = ccrs.Geodetic().transform_points(self.crs, cornerlons, cornerlats)

        # The following note in xESMF suggests that the arrays passed to ESMPy ought to
        # be fortran ordered:
        # https://xesmf.readthedocs.io/en/latest/internal_api.html#xesmf.backend.warn_f_contiguous
        # It is yet to be determined what effect this has on performance.
        truecenterlons = np.asfortranarray(truecenters[..., 0])
        truecenterlats = np.asfortranarray(truecenters[..., 1])
        truecornerlons = np.asfortranarray(truecorners[..., 0])
        truecornerlats = np.asfortranarray(truecorners[..., 1])

        info = (
            shape,
            truecenterlons,
            truecenterlats,
            truecornerlons,
            truecornerlats,
            self.circular,
            self.areas,
        )
        return info

    def _make_esmf_sdo(self):
        info = self._as_esmf_info()
        (
            shape,
            truecenterlons,
            truecenterlats,
            truecornerlons,
            truecornerlats,
            circular,
            areas,
        ) = info

        if circular:
            grid = esmpy.Grid(
                shape,
                pole_kind=[1, 1],
                num_peri_dims=1,
                periodic_dim=1,
                pole_dim=0,
            )
        else:
            grid = esmpy.Grid(shape, pole_kind=[1, 1])

        grid.add_coords(staggerloc=esmpy.StaggerLoc.CORNER)
        grid_corner_x = grid.get_coords(0, staggerloc=esmpy.StaggerLoc.CORNER)
        grid_corner_x[:] = truecornerlons
        grid_corner_y = grid.get_coords(1, staggerloc=esmpy.StaggerLoc.CORNER)
        grid_corner_y[:] = truecornerlats

        # Grid center points are added here, this is not necessary for
        # conservative area weighted regridding
        if self.center:
            grid.add_coords(staggerloc=esmpy.StaggerLoc.CENTER)
            grid_center_x = grid.get_coords(0, staggerloc=esmpy.StaggerLoc.CENTER)
            grid_center_x[:] = truecenterlons
            grid_center_y = grid.get_coords(1, staggerloc=esmpy.StaggerLoc.CENTER)
            grid_center_y[:] = truecenterlats

        def add_get_item(grid, **kwargs):
            grid.add_item(**kwargs)
            return grid.get_item(**kwargs)

        if self.mask is not None:
            grid_mask = add_get_item(
                grid, item=esmpy.GridItem.MASK, staggerloc=esmpy.StaggerLoc.CENTER
            )
            grid_mask[:] = self._refined_mask

        if areas is not None:
            grid_areas = add_get_item(
                grid, item=esmpy.GridItem.AREA, staggerloc=esmpy.StaggerLoc.CENTER
            )
            grid_areas[:] = areas.T

        return grid


class RefinedGridInfo(GridInfo):
    """
    Class for handling structured grids represented in :mod:`esmpy` in higher resolution.

    A specialised version of :class:`GridInfo`. Designed to provide higher
    accuracy conservative regridding for rectilinear grids, especially those with
    particularly large cells which may not be well represented by :mod:`esmpy`. This
    class differs from :class:`GridInfo` primarily in the way it represents itself
    as a :class:`~esmpy.api.field.Field` in :mod:`esmpy`. This :class:`~esmpy.api.field.Field`
    is designed to be a higher resolution version of the given grid and should
    contain enough information for area weighted regridding but may be
    inappropriate for other :mod:`esmpy` regridding schemes.

    """

    def __init__(
        self,
        lonbounds,
        latbounds,
        resolution=3,
        crs=None,
        mask=None,
    ):
        """
        Create a :class:`RefinedGridInfo` object describing the grid.

        Parameters
        ----------
        lonbounds : :obj:`~numpy.typing.ArrayLike`
            A 1D array or list describing the longitude bounds of the grid.
            Must be strictly increasing (for example, if a bound goes from
            170 to -170 consider transposing -170 to 190).
        latbounds : :obj:`~numpy.typing.ArrayLike`
            A 1D array or list describing the latitude bounds of the grid.
            Must be strictly increasing.
        resolution : int, default=400
            A number describing how many latitude slices each cell should
            be divided into when passing a higher resolution grid to ESMF.
        crs : :class:`cartopy.crs.CRS`, optional
            Describes how to interpret the
            above arguments. If ``None``, defaults to :class:`~cartopy.crs.Geodetic`.

        """
        # Convert bounds to numpy arrays where necessary.
        if not isinstance(lonbounds, np.ndarray):
            lonbounds = np.array(lonbounds)
        if not isinstance(latbounds, np.ndarray):
            latbounds = np.array(latbounds)

        # Ensure bounds are strictly increasing.
        if not np.all(lonbounds[:-1] < lonbounds[1:]):
            raise ValueError("The longitude bounds must be strictly increasing.")
        if not np.all(latbounds[:-1] < latbounds[1:]):
            raise ValueError("The latitude bounds must be strictly increasing.")

        self.resolution = resolution
        self.n_lons_orig = len(lonbounds) - 1
        self.n_lats_orig = len(latbounds) - 1

        # Create dummy lat/lon values
        lons = np.zeros(self.n_lons_orig)
        lats = np.zeros(self.n_lats_orig)
        super().__init__(lons, lats, lonbounds, latbounds, crs=crs, mask=mask)

        if self.n_lats_orig == 1 and np.allclose(latbounds, [-90, 90]):
            self._refined_latbounds = np.array([-90, 0, 90])
            self._refined_lonbounds = lonbounds
        else:
            self._refined_latbounds = latbounds
            self._refined_lonbounds = np.append(
                np.linspace(
                    lonbounds[:-1],
                    lonbounds[1:],
                    self.resolution,
                    endpoint=False,
                    axis=1,
                ).flatten(),
                lonbounds[-1],
            )
        self.lon_expansion = int(
            (len(self._refined_lonbounds) - 1) / (len(self.lonbounds) - 1)
        )
        self.lat_expansion = int(
            (len(self._refined_latbounds) - 1) / (len(self.latbounds) - 1)
        )

    @property
    def _refined_shape(self):
        """Return shape passed to ESMF."""
        return (
            self.n_lats_orig * self.lat_expansion,
            self.n_lons_orig * self.lon_expansion,
        )

    @property
    def _refined_mask(self):
        """Return mask passed to ESMF."""
        new_mask = np.broadcast_to(
            self.mask[:, np.newaxis, :, np.newaxis],
            [
                self.n_lats_orig,
                self.lat_expansion,
                self.n_lons_orig,
                self.lon_expansion,
            ],
        )
        new_mask = new_mask.reshape(self._refined_shape)
        return new_mask

    def _collapse_weights(self, is_tgt):
        """
        Return a matrix to collapse the weight matrix.

        The refined grid may contain more cells than the represented grid. When this is
        the case, the generated weight matrix will refer to too many points and will have
        to be collapsed. This is done by multiplying by this matrix, pre-multiplying when
        the target grid is represented and post multiplying when the source grid is
        represented.

        Parameters
        ----------
        is_tgt : bool
            True if the target field is being represented, False otherwise.
        """
        # The column indices represent each of the cells in the refined grid.
        column_indices = np.arange(self._refined_size)

        # The row indices represent the cells of the unrefined grid. These are broadcast
        # so that each row index coincides with all column indices of the refined cells
        # which the unrefined cell is split into.
        if self.lat_expansion > 1:
            # The latitudes are expanded only in the case where there is one latitude
            # bound from -90 to 90. In this case, there is no longitude expansion.
            row_indices = np.empty([self.n_lons_orig, self.lat_expansion])
            row_indices[:] = np.arange(self.n_lons_orig)[:, np.newaxis]
        else:
            # The row indices are broadcast across a dimension representing the expansion
            # of the longitude. Each row index is broadcast and flattened so that all the
            # row indices representing the unrefined cell match up with the column indices
            # representing the refined cells it is split into.
            row_indices = np.empty(
                [self.n_lons_orig, self.lon_expansion, self.n_lats_orig]
            )
            row_indices[:] = np.arange(self.n_lons_orig * self.n_lats_orig).reshape(
                [self.n_lons_orig, self.n_lats_orig]
            )[:, np.newaxis, :]
        row_indices = row_indices.flatten()
        matrix_shape = (self.size, self._refined_size)
        refinement_weights = scipy.sparse.csr_matrix(
            (
                np.ones(self._refined_size),
                (row_indices, column_indices),
            ),
            shape=matrix_shape,
        )
        if is_tgt:
            # When the RefinedGridInfo is the target of the regridder, we want to take
            # the average of the weights of each refined target cell. This is because
            # these weights represent the proportion of area of the target cells which
            # is covered by a given source cell. Since the refined cells are divided in
            # such a way that they have equal area, the weights for the unrefined cells
            # can be reconstructed by taking an average. This is done via matrix
            # multiplication, with the returned matrix pre-multiplying the weight matrix
            # so that it operates on the rows of the weight matrix (representing the
            # target cells). At this point the returned matrix consists of ones, so we
            # divided by the number of refined cells per unrefined cell.
            refinement_weights = refinement_weights / (
                self.lon_expansion * self.lat_expansion
            )
        else:
            # When the RefinedGridInfo is the source of the regridder, we want to take
            # the sum of the weights of each refined target cell. This is because those
            # weights represent the proportion of the area of a given target cell which
            # is covered by each refined source cell. The total proportion covered by
            # each unrefined source cell is then the sum of the weights from each of its
            # refined cells. This sum is done by matrix multiplication, the returned
            # matrix post-multiplying the weight matrix so that it operates on the columns
            # of the weight matrix (representing the source cells). In order for the
            # post-multiplication to work, the returned matrix must be transposed.
            refinement_weights = refinement_weights.T
        return refinement_weights



"""Unit tests for :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`."""

from cf_units import Unit
import numpy as np
import pytest

from esmf_regrid.schemes import ESMFAreaWeightedRegridder
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _curvilinear_cube,
    _grid_cube,
)


def test_dim_switching():
    """
    Test calling of :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that the regridder accepts a cube with dimensions in a different
    order than the cube which initialised it. Checks that dimension order is
    inherited from the cube in the calling function in both cases.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    regridder = ESMFAreaWeightedRegridder(src, tgt)
    unswitched_result = regridder(src)

    src_switched = src.copy()
    src_switched.transpose()
    switched_result = regridder(src_switched)

    assert unswitched_result.coord(dimensions=(0,)).standard_name == "latitude"
    assert unswitched_result.coord(dimensions=(1,)).standard_name == "longitude"
    assert switched_result.coord(dimensions=(0,)).standard_name == "longitude"
    assert switched_result.coord(dimensions=(1,)).standard_name == "latitude"


def test_differing_grids():
    """
    Test calling of :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that the regridder raises an error when given a cube with a different
    grid to the one it was initialised with.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src_init = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    n_lons_dif = 7
    src_dif_coord = _grid_cube(
        n_lons_dif, n_lats, lon_bounds, lat_bounds, circular=True
    )
    src_dif_circ = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=False)

    regridder = ESMFAreaWeightedRegridder(src_init, tgt)

    msg = "The given cube is not defined on the same source grid as this regridder."
    with pytest.raises(ValueError, match=msg):
        _ = regridder(src_dif_coord)
    with pytest.raises(ValueError, match=msg):
        _ = regridder(src_dif_circ)


def test_invalid_mdtol():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that an error is raised when mdtol is out of range.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    match = "Value for mdtol must be in range 0 - 1, got "
    with pytest.raises(ValueError, match=match):
        _ = ESMFAreaWeightedRegridder(src, tgt, mdtol=2)
    with pytest.raises(ValueError, match=match):
        _ = ESMFAreaWeightedRegridder(src, tgt, mdtol=-1)


def test_invalid_tgt_location():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that initialisation fails when tgt_location is not "face".
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    match = "For area weighted regridding, target location must be 'face'."
    with pytest.raises(ValueError, match=match):
        _ = ESMFAreaWeightedRegridder(src, tgt, tgt_location="node")


def test_curvilinear_equivalence():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that equivalent curvilinear and rectilinear coordinates give the same
    results.
    """
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)

    grid_to_grid = ESMFAreaWeightedRegridder(grid_src, grid_tgt)
    grid_to_curv = ESMFAreaWeightedRegridder(grid_src, curv_tgt)
    curv_to_grid = ESMFAreaWeightedRegridder(curv_src, grid_tgt)
    curv_to_curv = ESMFAreaWeightedRegridder(curv_src, curv_tgt)

    def extract_weights(regridder):
        return regridder.regridder.weight_matrix.todense()

    for regridder in [grid_to_curv, curv_to_grid, curv_to_curv]:
        assert np.allclose(extract_weights(grid_to_grid), extract_weights(regridder))


def test_curvilinear_and_rectilinear():
    """
    Test :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that a cube with both curvilinear and rectilinear coords still works.
    Checks that the DimCoords have priority over AuxCoords.
    """
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)

    src = curv_src.copy()
    grid_lat_src = grid_src.coord("latitude")
    grid_lat_src.standard_name = "grid_latitude"
    src.add_dim_coord(grid_lat_src, 0)
    grid_lon_src = grid_src.coord("longitude")
    grid_lon_src.standard_name = "grid_longitude"
    src.add_dim_coord(grid_lon_src, 1)

    tgt = curv_tgt.copy()
    grid_lat_tgt = grid_tgt.coord("latitude")
    grid_lat_tgt.standard_name = "grid_latitude"
    tgt.add_dim_coord(grid_lat_tgt, 0)
    grid_lon_tgt = grid_tgt.coord("longitude")
    grid_lon_tgt.standard_name = "grid_longitude"
    tgt.add_dim_coord(grid_lon_tgt, 1)

    # Change the AuxCoords to check that the DimCoords have priority.
    src.coord("latitude").bounds[:] = 0
    src.coord("longitude").bounds[:] = 0
    tgt.coord("latitude").bounds[:] = 0
    tgt.coord("longitude").bounds[:] = 0

    # Ensure the source data is all ones.
    src.data[:] = 1

    rg = ESMFAreaWeightedRegridder(src, tgt)
    result = rg(src)

    expected = grid_tgt.copy()
    # If the aux coords had been prioritised, expected.data would be a fully masked array.
    expected.data[:] = 1
    assert expected == result
    assert not np.ma.is_masked(result)


def test_unit_equivalence():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that equivalent coordinates in degrees and radians give the same results.
    """
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    lon_rad_bounds = (-np.pi, np.pi)
    lat_rad_bounds = (-np.pi / 2, np.pi / 2)

    def rad_coords(cube):
        cube.coord("latitude").units = Unit("radians")
        cube.coord("longitude").units = Unit("radians")

    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_src_rad = _grid_cube(
        n_lons_src, n_lats_src, lon_rad_bounds, lat_rad_bounds, circular=True
    )
    rad_coords(grid_src_rad)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    grid_tgt_rad = _grid_cube(
        n_lons_tgt, n_lats_tgt, lon_rad_bounds, lat_rad_bounds, circular=True
    )
    rad_coords(grid_tgt_rad)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_src_rad = _curvilinear_cube(
        n_lons_src, n_lats_src, lon_rad_bounds, lat_rad_bounds
    )
    rad_coords(curv_src_rad)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)
    curv_tgt_rad = _curvilinear_cube(
        n_lons_tgt, n_lats_tgt, lon_rad_bounds, lat_rad_bounds
    )
    rad_coords(curv_tgt_rad)

    grid_to_grid = ESMFAreaWeightedRegridder(grid_src, grid_tgt)
    grid_rad_to_grid = ESMFAreaWeightedRegridder(grid_src_rad, grid_tgt)
    grid_rad_to_curv = ESMFAreaWeightedRegridder(grid_src_rad, curv_tgt)
    curv_to_grid_rad = ESMFAreaWeightedRegridder(curv_src, grid_tgt_rad)
    curv_rad_to_grid = ESMFAreaWeightedRegridder(curv_src_rad, grid_tgt)
    curv_to_curv_rad = ESMFAreaWeightedRegridder(curv_src, curv_tgt_rad)

    def extract_weights(regridder):
        return regridder.regridder.weight_matrix.todense()

    for regridder in [
        grid_rad_to_grid,
        grid_rad_to_curv,
        curv_to_grid_rad,
        curv_rad_to_grid,
        curv_to_curv_rad,
    ]:
        assert np.allclose(extract_weights(grid_to_grid), extract_weights(regridder))


def test_masks():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that the `use_src_mask` and `use_tgt_mask` keywords work properly.
    """
    src = _curvilinear_cube(7, 6, [-180, 180], [-90, 90])
    tgt = _curvilinear_cube(6, 7, [-180, 180], [-90, 90])

    # Make src and tgt discontiguous at (0, 0)
    src_mask = np.zeros([6, 7], dtype=bool)
    src_mask[0, 0] = True
    src.data = np.ma.array(src.data, mask=src_mask)
    src_discontiguous = src.copy()
    src_discontiguous.coord("latitude").bounds[0, 0] = 0
    src_discontiguous.coord("longitude").bounds[0, 0] = 0

    tgt_mask = np.zeros([7, 6], dtype=bool)
    tgt_mask[0, 0] = True
    tgt.data = np.ma.array(tgt.data, mask=tgt_mask)
    tgt_discontiguous = tgt.copy()
    tgt_discontiguous.coord("latitude").bounds[0, 0] = 0
    tgt_discontiguous.coord("longitude").bounds[0, 0] = 0

    rg_src_masked = ESMFAreaWeightedRegridder(src_discontiguous, tgt, use_src_mask=True)
    rg_tgt_masked = ESMFAreaWeightedRegridder(src, tgt_discontiguous, use_tgt_mask=True)
    rg_unmasked = ESMFAreaWeightedRegridder(src, tgt)

    weights_src_masked = rg_src_masked.regridder.weight_matrix
    weights_tgt_masked = rg_tgt_masked.regridder.weight_matrix
    weights_unmasked = rg_unmasked.regridder.weight_matrix

    # Check there are no weights associated with the masked point.
    assert weights_src_masked[:, 0].nnz == 0
    assert weights_tgt_masked[0].nnz == 0

    # Check all other weights are correct.
    assert np.allclose(
        weights_src_masked[:, 1:].todense(), weights_unmasked[:, 1:].todense()
    )
    assert np.allclose(weights_tgt_masked[1:].todense(), weights_unmasked[1:].todense())


def test_resolution():
    """
    Test calling of :class:`esmf_regrid.schemes.ESMFAreaWeightedRegridder`.

    Checks that the regridder accepts resolution arguments.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    src_resolution = 3
    tgt_resolution = 4

    regridder = ESMFAreaWeightedRegridder(
        src, tgt, src_resolution=src_resolution, tgt_resolution=tgt_resolution
    )
    assert regridder.src_resolution == src_resolution
    assert regridder.regridder.src.resolution == src_resolution
    assert regridder.tgt_resolution == tgt_resolution
    assert regridder.regridder.tgt.resolution == tgt_resolution



"""Unit tests for :class:`esmf_regrid.schemes.ESMFNearestRegridder`."""

from cf_units import Unit
import numpy as np
import pytest

from esmf_regrid.schemes import ESMFNearestRegridder
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _curvilinear_cube,
    _grid_cube,
)
from esmf_regrid.tests.unit.schemes.test__mesh_to_MeshInfo import (
    _gridlike_mesh_cube,
)


def test_dim_switching():
    """
    Test calling of :class:`esmf_regrid.schemes.ESMFNearestRegridder`.

    Checks that the regridder accepts a cube with dimensions in a different
    order than the cube which initialised it. Checks that dimension order is
    inherited from the cube in the calling function in both cases.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    regridder = ESMFNearestRegridder(src, tgt)
    unswitched_result = regridder(src)

    src_switched = src.copy()
    src_switched.transpose()
    switched_result = regridder(src_switched)

    assert unswitched_result.coord(dimensions=(0,)).standard_name == "latitude"
    assert unswitched_result.coord(dimensions=(1,)).standard_name == "longitude"
    assert switched_result.coord(dimensions=(0,)).standard_name == "longitude"
    assert switched_result.coord(dimensions=(1,)).standard_name == "latitude"


def test_differing_grids():
    """
    Test calling of :class:`esmf_regrid.schemes.ESMFNearestRegridder`.

    Checks that the regridder raises an error when given a cube with a different
    grid to the one it was initialised with.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src_init = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    n_lons_dif = 7
    src_dif_coord = _grid_cube(
        n_lons_dif, n_lats, lon_bounds, lat_bounds, circular=True
    )
    src_dif_circ = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=False)

    regridder = ESMFNearestRegridder(src_init, tgt)

    msg = "The given cube is not defined on the same source grid as this regridder."
    with pytest.raises(ValueError, match=msg):
        _ = regridder(src_dif_coord)
    with pytest.raises(ValueError, match=msg):
        _ = regridder(src_dif_circ)


def test_curvilinear_equivalence():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFNearestRegridder`.

    Checks that equivalent curvilinear and rectilinear coordinates give the same
    results.
    """
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)

    grid_to_grid = ESMFNearestRegridder(grid_src, grid_tgt)
    grid_to_curv = ESMFNearestRegridder(grid_src, curv_tgt)
    curv_to_grid = ESMFNearestRegridder(curv_src, grid_tgt)
    curv_to_curv = ESMFNearestRegridder(curv_src, curv_tgt)

    def extract_weights(regridder):
        return regridder.regridder.weight_matrix.todense()

    for regridder in [grid_to_curv, curv_to_grid, curv_to_curv]:
        assert np.allclose(extract_weights(grid_to_grid), extract_weights(regridder))


def test_curvilinear_and_rectilinear():
    """
    Test :class:`esmf_regrid.schemes.ESMFNearestRegridder`.

    Checks that a cube with both curvilinear and rectilinear coords still works.
    Checks that the DimCoords have priority over AuxCoords.
    """
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)

    src = curv_src.copy()
    grid_lat_src = grid_src.coord("latitude")
    grid_lat_src.standard_name = "grid_latitude"
    src.add_dim_coord(grid_lat_src, 0)
    grid_lon_src = grid_src.coord("longitude")
    grid_lon_src.standard_name = "grid_longitude"
    src.add_dim_coord(grid_lon_src, 1)

    tgt = curv_tgt.copy()
    grid_lat_tgt = grid_tgt.coord("latitude")
    grid_lat_tgt.standard_name = "grid_latitude"
    tgt.add_dim_coord(grid_lat_tgt, 0)
    grid_lon_tgt = grid_tgt.coord("longitude")
    grid_lon_tgt.standard_name = "grid_longitude"
    tgt.add_dim_coord(grid_lon_tgt, 1)

    # Change the AuxCoords to check that the DimCoords have priority.
    src.coord("latitude").bounds[:] = 0
    src.coord("longitude").bounds[:] = 0
    tgt.coord("latitude").bounds[:] = 0
    tgt.coord("longitude").bounds[:] = 0

    # Ensure the source data is all ones.
    src.data[:] = 1

    rg = ESMFNearestRegridder(src, tgt)
    result = rg(src)

    expected = grid_tgt.copy()
    # If the aux coords had been prioritised, expected.data would be a fully masked array.
    expected.data[:] = 1
    assert expected == result
    assert not np.ma.is_masked(result)


def test_unit_equivalence():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFNearestRegridder`.

    Checks that equivalent coordinates in degrees and radians give the same results.
    """
    # While this test has been copied from test_ESMFBilinerRegridder.py, a slight
    # change has been made to the parameter n_lons_src.
    # It should be noted that before this change this test was failing.
    # This is due to the fact that when two source points are equidistant from a
    # target point, rounding floating point differences due to unit conversion
    # would have an effect of the result.
    # These parameters have been tweaked so that no target point is equidistant from
    # two nearest source points.
    n_lons_src = 5
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    lon_rad_bounds = (-np.pi, np.pi)
    lat_rad_bounds = (-np.pi / 2, np.pi / 2)

    def rad_coords(cube):
        cube.coord("latitude").units = Unit("radians")
        cube.coord("longitude").units = Unit("radians")

    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_src_rad = _grid_cube(
        n_lons_src, n_lats_src, lon_rad_bounds, lat_rad_bounds, circular=True
    )
    rad_coords(grid_src_rad)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    grid_tgt_rad = _grid_cube(
        n_lons_tgt, n_lats_tgt, lon_rad_bounds, lat_rad_bounds, circular=True
    )
    rad_coords(grid_tgt_rad)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_src_rad = _curvilinear_cube(
        n_lons_src, n_lats_src, lon_rad_bounds, lat_rad_bounds
    )
    rad_coords(curv_src_rad)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)
    curv_tgt_rad = _curvilinear_cube(
        n_lons_tgt, n_lats_tgt, lon_rad_bounds, lat_rad_bounds
    )
    rad_coords(curv_tgt_rad)

    grid_to_grid = ESMFNearestRegridder(grid_src, grid_tgt)
    grid_rad_to_grid = ESMFNearestRegridder(grid_src_rad, grid_tgt)
    grid_rad_to_curv = ESMFNearestRegridder(grid_src_rad, curv_tgt)
    curv_to_grid_rad = ESMFNearestRegridder(curv_src, grid_tgt_rad)
    curv_rad_to_grid = ESMFNearestRegridder(curv_src_rad, grid_tgt)
    curv_to_curv_rad = ESMFNearestRegridder(curv_src, curv_tgt_rad)

    def extract_weights(regridder):
        return regridder.regridder.weight_matrix.todense()

    for regridder in [
        grid_rad_to_grid,
        grid_rad_to_curv,
        curv_to_grid_rad,
        curv_rad_to_grid,
        curv_to_curv_rad,
    ]:
        assert np.allclose(extract_weights(grid_to_grid), extract_weights(regridder))


def test_masks():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFNearestRegridder`.

    Checks that the `use_src_mask` and `use_tgt_mask` keywords work properly.
    """
    src = _curvilinear_cube(7, 6, [-180, 180], [-90, 90])
    tgt = _curvilinear_cube(6, 7, [-180, 180], [-90, 90])

    # Make src and tgt discontiguous at (0, 0)
    src_mask = np.zeros([6, 7], dtype=bool)
    src_mask[0, 0] = True
    src.data = np.ma.array(src.data, mask=src_mask)
    src_discontiguous = src.copy()
    src_discontiguous.coord("latitude").bounds[0, 0] = 0
    src_discontiguous.coord("longitude").bounds[0, 0] = 0

    tgt_mask = np.zeros([7, 6], dtype=bool)
    tgt_mask[0, 0] = True
    tgt.data = np.ma.array(tgt.data, mask=tgt_mask)
    tgt_discontiguous = tgt.copy()
    tgt_discontiguous.coord("latitude").bounds[0, 0] = 0
    tgt_discontiguous.coord("longitude").bounds[0, 0] = 0

    rg_src_masked = ESMFNearestRegridder(src_discontiguous, tgt, use_src_mask=True)
    rg_tgt_masked = ESMFNearestRegridder(src, tgt_discontiguous, use_tgt_mask=True)
    rg_unmasked = ESMFNearestRegridder(src, tgt)

    weights_src_masked = rg_src_masked.regridder.weight_matrix
    weights_tgt_masked = rg_tgt_masked.regridder.weight_matrix
    weights_unmasked = rg_unmasked.regridder.weight_matrix

    # Check there are no weights associated with the masked point.
    assert weights_src_masked[:, 0].nnz == 0
    assert weights_tgt_masked[0].nnz == 0

    # Check other weights are correct. Note that unique to NEAREST_DTOS, masking a source
    # point causes the next nearest source point to gain weights in the weight matrix.
    # because of this, we ignore the row associated with that target point and check
    # the rest of the weights matrix.
    assert np.allclose(weights_src_masked[1:].todense(), weights_unmasked[1:].todense())
    assert np.allclose(weights_tgt_masked[1:].todense(), weights_unmasked[1:].todense())


@pytest.mark.parametrize(
    "src_type,tgt_type",
    [
        ("grid", "grid"),
        ("grid", "mesh"),
        ("mesh", "grid"),
        ("grid", "curv"),
        ("curv", "grid"),
    ],
)
def test_regrid_data(src_type, tgt_type):
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFNearestRegridder`.

    Checks that regridding mathematics behaves in an expected way.
    """
    # Create two similar grids so that source data and expected result
    # data ought to look similar by visual inspection.
    if src_type == "grid":
        src = _grid_cube(5, 4, [-180, 180], [-90, 90], circular=True)
    elif src_type == "mesh":
        src = _gridlike_mesh_cube(5, 4)
    elif src_type == "curv":
        src = _curvilinear_cube(5, 4, [-180, 180], [-90, 90])

    if tgt_type == "grid":
        tgt = _grid_cube(4, 5, [-180, 180], [-90, 90], circular=True)
    elif tgt_type == "mesh":
        tgt = _gridlike_mesh_cube(4, 5)
    elif tgt_type == "curv":
        tgt = _curvilinear_cube(4, 5, [-180, 180], [-90, 90])

    if src_type == "mesh":
        src.data = np.arange(20)
    else:
        src_data = np.arange(20).reshape([4, 5])
        src.data = src_data

    rg = ESMFNearestRegridder(src, tgt)

    # when two source points are equidistant from a
    # target point, the chosen source point is dependent on the index which ESMF
    # gives that point. This decision is described by ESMF to be arbitrary, but
    # ought to be consistent when dealing with the same precise grid.
    expected_data = np.array(
        [
            [0, 1, 3, 4],
            [5, 6, 8, 9],
            [5, 6, 8, 9],
            [10, 11, 13, 14],
            [15, 16, 18, 19],
        ]
    )
    if tgt_type == "mesh":
        expected_data = expected_data.flatten()

    result = rg(src)
    np.testing.assert_allclose(expected_data, result.data)



"""Unit tests for miscellaneous helper functions in `esmf_regrid.experimental.unstructured_scheme`."""

import iris
from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube
import numpy as np

from esmf_regrid.schemes import _create_cube


def test_create_cube_2D():
    """Test creation of 2D output grid."""
    data = np.ones([2, 3])

    # Create a source cube with metadata and scalar coords
    src_cube = Cube(np.zeros(5))
    src_cube.units = "K"
    src_cube.attributes = {"a": 1}
    src_cube.standard_name = "air_temperature"
    scalar_height = AuxCoord([5], units="m", standard_name="height")
    scalar_time = DimCoord([10], units="s", standard_name="time")
    src_cube.add_aux_coord(scalar_height)
    src_cube.add_aux_coord(scalar_time)

    mesh_dim = 0

    grid_x = DimCoord(np.arange(3), standard_name="longitude")
    grid_y = DimCoord(np.arange(2), standard_name="latitude")

    cube = _create_cube(data, src_cube, (mesh_dim,), (grid_x, grid_y), 2)
    src_metadata = src_cube.metadata

    expected_cube = Cube(data)
    expected_cube.metadata = src_metadata
    expected_cube.add_dim_coord(grid_x, 1)
    expected_cube.add_dim_coord(grid_y, 0)
    expected_cube.add_aux_coord(scalar_height)
    expected_cube.add_aux_coord(scalar_time)
    assert expected_cube == cube


def test_create_cube_4D():
    """Test creation of 2D output grid."""
    data = np.ones([4, 2, 3, 5])

    # Create a source cube with metadata and scalar coords
    src_cube = Cube(np.zeros([4, 5, 5]))
    src_cube.units = "K"
    src_cube.attributes = {"a": 1}
    src_cube.standard_name = "air_temperature"
    scalar_height = AuxCoord([5], units="m", standard_name="height")
    scalar_time = DimCoord([10], units="s", standard_name="time")
    src_cube.add_aux_coord(scalar_height)
    src_cube.add_aux_coord(scalar_time)
    first_coord = DimCoord(np.arange(4), standard_name="air_pressure")
    src_cube.add_dim_coord(first_coord, 0)
    last_coord = AuxCoord(np.arange(5), long_name="last_coord")
    src_cube.add_aux_coord(last_coord, 2)
    multidim_coord = AuxCoord(np.ones([4, 5]), long_name="2d_coord")
    src_cube.add_aux_coord(multidim_coord, (0, 2))
    ignored_coord = AuxCoord(np.arange(5), long_name="ignore")
    src_cube.add_aux_coord(ignored_coord, 1)

    mesh_dim = 1

    grid_x = iris.coords.DimCoord(np.arange(3), standard_name="longitude")
    grid_y = iris.coords.DimCoord(np.arange(2), standard_name="latitude")

    cube = _create_cube(data, src_cube, (mesh_dim,), (grid_x, grid_y), 2)
    src_metadata = src_cube.metadata

    expected_cube = iris.cube.Cube(data)
    expected_cube.metadata = src_metadata
    expected_cube.add_dim_coord(grid_x, 2)
    expected_cube.add_dim_coord(grid_y, 1)
    expected_cube.add_dim_coord(first_coord, 0)
    expected_cube.add_aux_coord(last_coord, 3)
    expected_cube.add_aux_coord(multidim_coord, (0, 3))
    expected_cube.add_aux_coord(scalar_height)
    expected_cube.add_aux_coord(scalar_time)
    assert expected_cube == cube



"""Provides an iris interface for unstructured regridding."""

from iris.experimental.ugrid import Mesh

from esmf_regrid import check_method, Constants
from esmf_regrid.schemes import (
    _ESMFRegridder,
    _get_mask,
    _regrid_rectilinear_to_unstructured__perform,
    _regrid_rectilinear_to_unstructured__prepare,
    _regrid_unstructured_to_rectilinear__perform,
    _regrid_unstructured_to_rectilinear__prepare,
    _regrid_unstructured_to_unstructured__perform,
    _regrid_unstructured_to_unstructured__prepare,
)


def regrid_unstructured_to_rectilinear(
    src_cube,
    grid_cube,
    mdtol=0,
    method=Constants.Method.CONSERVATIVE,
    tgt_resolution=None,
    use_src_mask=False,
    use_tgt_mask=False,
):
    r"""
    Regrid unstructured :class:`~iris.cube.Cube` onto rectilinear grid.

    Return a new :class:`~iris.cube.Cube` with :attr:`~iris.cube.Cube.data`
    values calculated using weights generated by :mod:`esmpy` to give the weighted
    mean of :attr:`~iris.cube.Cube.data` values from ``src_cube`` regridded onto the
    horizontal grid of ``grid_cube``. The dimension on the :class:`~iris.cube.Cube`
    belonging to the :attr:`~iris.cube.Cube.mesh`
    will be replaced by the two dimensions associated with the grid.
    This function requires that the horizontal dimension of ``src_cube`` is
    described by a 2D mesh with data located on the faces of that mesh
    for conservative regridding and located on either faces or nodes for
    bilinear regridding.
    This function allows the horizontal grid of ``grid_cube`` to be either
    rectilinear or curvilinear (i.e. expressed in terms of two orthogonal
    1D coordinates or via a pair of 2D coordinates).
    This function also requires that the :class:`~iris.coords.Coord`\\ s describing the
    horizontal grid have :attr:`~iris.coords.Coord.bounds`.

    Parameters
    ----------
    src_cube : :class:`iris.cube.Cube`
        An unstructured instance of :class:`~iris.cube.Cube` that supplies the data,
        metadata and coordinates.
    grid_cube : :class:`iris.cube.Cube`
        An instance of :class:`~iris.cube.Cube` that supplies the desired
        horizontal grid definition.
    mdtol : float, default=0
        Tolerance of missing data. The value returned in each element of the
        returned :class:`~iris.cube.Cube`\\ 's :attr:`~iris.cube.Cube.data`
        array will be masked if the fraction of masked
        data in the overlapping cells of ``src_cube`` exceeds ``mdtol``. This
        fraction is calculated based on the area of masked cells within each
        target cell. ``mdtol=0`` means no missing data is tolerated while ``mdtol=1``
        will mean the resulting element will be masked if and only if all the
        overlapping cells of ``src_cube`` are masked.
    method : :class:`Constants.Method`, default=Constants.Method.CONSERVATIVE
        The method used to calculate weights.
    tgt_resolution : int, optional
        If present, represents the amount of latitude slices per cell
        given to ESMF for calculation.
    use_src_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
        Either an array representing the cells in the source to ignore, or else
        a boolean value. If True, this array is taken from the mask on the data
        in ``src_cube``. If False, no mask will be taken and all points will
        be used in weights calculation.
    use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
        Either an array representing the cells in the target to ignore, or else
        a boolean value. If True, this array is taken from the mask on the data
        in ``grid_cube``. If False, no mask will be taken and all points
        will be used in weights calculation.

    Returns
    -------
    :class:`iris.cube.Cube`
        A new :class:`~iris.cube.Cube` instance.

    """
    if src_cube.mesh is None:
        raise ValueError("src_cube has no mesh.")
    src_mask = _get_mask(src_cube, use_src_mask)
    tgt_mask = _get_mask(grid_cube, use_tgt_mask)
    method = check_method(method)

    regrid_info = _regrid_unstructured_to_rectilinear__prepare(
        src_cube,
        grid_cube,
        method=method,
        tgt_resolution=tgt_resolution,
        src_mask=src_mask,
        tgt_mask=tgt_mask,
    )
    result = _regrid_unstructured_to_rectilinear__perform(src_cube, regrid_info, mdtol)
    return result


class MeshToGridESMFRegridder(_ESMFRegridder):
    r"""Regridder class for unstructured to rectilinear :class:`~iris.cube.Cube`\\ s."""

    def __init__(
        self,
        src,
        tgt,
        mdtol=None,
        method=Constants.Method.CONSERVATIVE,
        precomputed_weights=None,
        tgt_resolution=None,
        use_src_mask=False,
        use_tgt_mask=False,
    ):
        """
        Create regridder for conversions between source mesh and target grid.

        Parameters
        ----------
        src_mesh_cube : :class:`iris.cube.Cube`
            The unstructured :class:`~iris.cube.Cube` providing the source mesh.
        target_grid_cube : :class:`iris.cube.Cube`
            The :class:`~iris.cube.Cube` providing the target grid.
        mdtol : float, optional
            Tolerance of missing data. The value returned in each element of
            the returned array will be masked if the fraction of masked data
            exceeds ``mdtol``. ``mdtol=0`` means no missing data is tolerated while
            ``mdtol=1`` will mean the resulting element will be masked if and only
            if all the contributing elements of data are masked. Defaults to 1
            for conservative regregridding and 0 for bilinear regridding.
        method : :class:`Constants.Method`, default=Constants.Method.CONSERVATIVE
            The method used to calculate weights.
        precomputed_weights : :class:`scipy.sparse.spmatrix`, optional
            If ``None``, :mod:`esmpy` will be used to
            calculate regridding weights. Otherwise, :mod:`esmpy` will be bypassed
            and ``precomputed_weights`` will be used as the regridding weights.
        tgt_resolution : int, optional
            If present, represents the amount of latitude slices per cell
            given to ESMF for calculation. If ``tgt_resolution`` is set, ``tgt``
            must have strictly increasing bounds (bounds may be transposed plus or
            minus 360 degrees to make the bounds strictly increasing).
        use_src_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
            Either an array representing the cells in the source to ignore, or else
            a boolean value. If True, this array is taken from the mask on the data
            in ``src``. If False, no mask will be taken and all points will
            be used in weights calculation.
        use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
            Either an array representing the cells in the target to ignore, or else
            a boolean value. If True, this array is taken from the mask on the data
            in ``tgt``. If False, no mask will be taken and all points
            will be used in weights calculation.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src``
            or ``tgt`` respectively are not constant over non-horizontal dimensions.


        """
        if src.mesh is None:
            raise ValueError("src has no mesh.")
        super().__init__(
            src,
            tgt,
            method,
            mdtol=mdtol,
            precomputed_weights=precomputed_weights,
            tgt_resolution=tgt_resolution,
            use_src_mask=use_src_mask,
            use_tgt_mask=use_tgt_mask,
        )
        self.resolution = tgt_resolution
        self.mesh, self.location = self._src
        self.grid_x, self.grid_y = self._tgt


def regrid_rectilinear_to_unstructured(
    src_cube,
    mesh_cube,
    mdtol=0,
    method=Constants.Method.CONSERVATIVE,
    src_resolution=None,
    use_src_mask=False,
    use_tgt_mask=False,
):
    r"""
    Regrid rectilinear :class:`~iris.cube.Cube` onto unstructured mesh.

    Return a new :class:`~iris.cube.Cube` with :attr:`~iris.cube.Cube.data`
    values calculated using weights generated by :mod:`esmpy` to give the weighted
    mean of :attr:`~iris.cube.Cube.data` values from ``src_cube`` regridded onto the
    horizontal mesh of ``mesh_cube``. The dimensions on the :class:`~iris.cube.Cube` associated
    with the grid will be replaced by a dimension associated with the
    :attr:`~iris.cube.Cube.mesh`.
    That dimension will be the first of the grid dimensions, whether
    it is associated with the ``x`` or ``y`` coordinate. Since two dimensions are
    being replaced by one, coordinates associated with dimensions after
    the grid will become associated with dimensions one lower.
    This function requires that the horizontal dimension of ``mesh_cube`` is
    described by a 2D mesh with data located on the faces of that mesh
    for conservative regridding and located on either faces or nodes for
    bilinear regridding.
    This function allows the horizontal grid of ``grid_cube`` to be either
    rectilinear or curvilinear (i.e. expressed in terms of two orthogonal
    1D coordinates or via a pair of 2D coordinates).
    This function also requires that the :class:`~iris.coords.Coord`\\ s describing the
    horizontal grid have :attr:`~iris.coords.Coord.bounds`.

    Parameters
    ----------
    src_cube : :class:`iris.cube.Cube`
        A rectilinear instance of :class:`~iris.cube.Cube` that supplies the data,
        metadata and coordinates.
    mesh_cube : :class:`iris.cube.Cube`
        An unstructured instance of :class:`~iris.cube.Cube` that supplies the desired
        horizontal mesh definition.
    mdtol : float, default=0
        Tolerance of missing data. The value returned in each element of the
        returned :class:`~iris.cube.Cube`\\ 's :attr:`~iris.cube.Cube.data` array
        will be masked if the fraction of masked
        data in the overlapping cells of the source cube exceeds ``mdtol``. This
        fraction is calculated based on the area of masked cells within each
        target cell. ``mdtol=0`` means no missing data is tolerated while ``mdtol=1``
        will mean the resulting element will be masked if and only if all the
        overlapping cells of the ``src_cube`` are masked.
    method : :class:`Constants.Method`, default=Constants.Method.CONSERVATIVE
        The method used to calculate weights.
    src_resolution : int, optional
        If present, represents the amount of latitude slices per cell
        given to ESMF for calculation.
    use_src_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
        Either an array representing the cells in the source to ignore, or else
        a boolean value. If True, this array is taken from the mask on the data
        in ``src_cube``. If False, no mask will be taken and all points will
        be used in weights calculation.
    use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
        Either an array representing the cells in the target to ignore, or else
        a boolean value. If True, this array is taken from the mask on the data
        in ``grid_cube``. If False, no mask will be taken and all points
        will be used in weights calculation.

    Returns
    -------
    :class:`iris.cube.Cube`
        A new :class:`~iris.cube.Cube` instance.

    """
    if mesh_cube.mesh is None:
        raise ValueError("mesh_cube has no mesh.")
    src_mask = _get_mask(src_cube, use_src_mask)
    tgt_mask = _get_mask(mesh_cube, use_tgt_mask)
    method = check_method(method)

    regrid_info = _regrid_rectilinear_to_unstructured__prepare(
        src_cube,
        mesh_cube,
        method=method,
        src_resolution=src_resolution,
        src_mask=src_mask,
        tgt_mask=tgt_mask,
    )
    result = _regrid_rectilinear_to_unstructured__perform(src_cube, regrid_info, mdtol)
    return result


class GridToMeshESMFRegridder(_ESMFRegridder):
    r"""Regridder class for rectilinear to unstructured :class:`~iris.cube.Cube`\\ s."""

    def __init__(
        self,
        src,
        tgt,
        mdtol=None,
        method=Constants.Method.CONSERVATIVE,
        precomputed_weights=None,
        src_resolution=None,
        use_src_mask=False,
        use_tgt_mask=False,
        tgt_location=None,
    ):
        """
        Create regridder for conversions between source grid and target mesh.

        Parameters
        ----------
        src : :class:`iris.cube.Cube`
            The rectilinear :class:`~iris.cube.Cube` cube providing the source grid.
        tgt : :class:`iris.cube.Cube` or :class:`iris.experimental.ugrid.Mesh`
            The unstructured :class:`~iris.cube.Cube`or
            :class:`~iris.experimental.ugrid.Mesh` providing the target mesh.
        mdtol : float, optional
            Tolerance of missing data. The value returned in each element of
            the returned array will be masked if the fraction of masked data
            exceeds ``mdtol``. ``mdtol=0`` means no missing data is tolerated while
            ``mdtol=1`` will mean the resulting element will be masked if and only
            if all the contributing elements of data are masked. Defaults to 1
            for conservative regregridding and 0 for bilinear regridding.
        method : :class:`Constants.Method`, default=Constants.Method.CONSERVATIVE
            The method used to calculate weights.
        precomputed_weights : :class:`scipy.sparse.spmatrix`, optional
            If ``None``, :mod:`esmpy` will be used to
            calculate regridding weights. Otherwise, :mod:`esmpy` will be bypassed
            and ``precomputed_weights`` will be used as the regridding weights.
        src_resolution : int, optional
            If present, represents the amount of latitude slices per cell
            given to ESMF for calculation. If ``src_resolution`` is set, ``src``
            must have strictly increasing bounds (bounds may be transposed plus or
            minus 360 degrees to make the bounds strictly increasing).
        use_src_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
            Either an array representing the cells in the source to ignore, or else
            a boolean value. If True, this array is taken from the mask on the data
            in ``src``. If False, no mask will be taken and all points will
            be used in weights calculation.
        use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
            Either an array representing the cells in the target to ignore, or else
            a boolean value. If True, this array is taken from the mask on the data
            in ``tgt``. If False, no mask will be taken and all points
            will be used in weights calculation.
        tgt_location : str or None, default=None
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src``
            or ``tgt`` respectively are not constant over non-horizontal dimensions.

        """
        if not isinstance(tgt, Mesh) and tgt.mesh is None:
            raise ValueError("tgt has no mesh.")
        super().__init__(
            src,
            tgt,
            method,
            mdtol=mdtol,
            precomputed_weights=precomputed_weights,
            src_resolution=src_resolution,
            use_src_mask=use_src_mask,
            use_tgt_mask=use_tgt_mask,
            tgt_location=tgt_location,
        )
        self.resolution = src_resolution
        self.mesh, self.location = self._tgt
        self.grid_x, self.grid_y = self._src


def regrid_unstructured_to_unstructured(
    src_mesh_cube,
    tgt_mesh_cube,
    mdtol=0,
    method=Constants.Method.CONSERVATIVE,
    use_src_mask=False,
    use_tgt_mask=False,
):
    r"""
    Regrid rectilinear :class:`~iris.cube.Cube` onto unstructured mesh.

    Return a new :class:`~iris.cube.Cube` with :attr:`~iris.cube.Cube.data`
    values calculated using weights generated by :mod:`esmpy` to give the weighted
    mean of :attr:`~iris.cube.Cube.data` values from ``src_mesh_cube`` regridded onto the
    horizontal mesh of ``tgt_mesh_cube``. The resulting cube will have the same
    ``mesh_dim`` as ``src_mesh_cube``.

    Parameters
    ----------
    src_mesh_cube : :class:`iris.cube.Cube`
        A unstructured instance of :class:`~iris.cube.Cube` that supplies the data,
        metadata and coordinates.
    tgt_mesh_cube : :class:`iris.cube.Cube`
        An unstructured instance of :class:`~iris.cube.Cube` that supplies the desired
        horizontal mesh definition.
    mdtol : float, default=0
        Tolerance of missing data. The value returned in each element of the
        returned :class:`~iris.cube.Cube`\\ 's :attr:`~iris.cube.Cube.data` array
        will be masked if the fraction of masked
        data in the overlapping cells of the source cube exceeds ``mdtol``. This
        fraction is calculated based on the area of masked cells within each
        target cell. ``mdtol=0`` means no missing data is tolerated while ``mdtol=1``
        will mean the resulting element will be masked if and only if all the
        overlapping cells of the ``src_cube`` are masked.
    method : :class:`Constants.Method`, default=Constants.Method.CONSERVATIVE
            The method used to calculate weights.
    use_src_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
        Either an array representing the cells in the source to ignore, or else
        a boolean value. If True, this array is taken from the mask on the data
        in ``src_cube``. If False, no mask will be taken and all points will
        be used in weights calculation.
    use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
        Either an array representing the cells in the target to ignore, or else
        a boolean value. If True, this array is taken from the mask on the data
        in ``grid_cube``. If False, no mask will be taken and all points
        will be used in weights calculation.

    Returns
    -------
    :class:`iris.cube.Cube`
        A new :class:`~iris.cube.Cube` instance.

    """
    method = check_method(method)
    if tgt_mesh_cube.mesh is None:
        raise ValueError("mesh_cube has no mesh.")
    src_mask = _get_mask(src_mesh_cube, use_src_mask)
    tgt_mask = _get_mask(tgt_mesh_cube, use_tgt_mask)

    regrid_info = _regrid_unstructured_to_unstructured__prepare(
        src_mesh_cube,
        tgt_mesh_cube,
        method=method,
        src_mask=src_mask,
        tgt_mask=tgt_mask,
    )
    result = _regrid_unstructured_to_unstructured__perform(
        src_mesh_cube, regrid_info, mdtol
    )
    return result



"""Common testing infrastructure."""

import pathlib

import numpy as np


# Base directory of the test results.
_RESULT_PATH = pathlib.Path(__file__).parent.resolve() / "results"


def get_result_path(relative_path, unit=True):
    """
    Form the absolute path to the results test file.

    Parameters
    ----------
    relative_path : str, path, iterable str/path
        The relative path to the target test file.
    unit : bool, optional
        Specify whether the `relative_path` is for a unit test.
        Default is True.

    Returns
    -------
    Path
        The absolute result path.

    """
    if isinstance(relative_path, str):
        relative_path = pathlib.Path(relative_path)

    if not isinstance(relative_path, pathlib.PurePath):
        relative_path = pathlib.Path(*relative_path)

    if unit:
        relative_path = pathlib.Path("unit") / relative_path

    result = _RESULT_PATH / relative_path

    return result.resolve(strict=True)


def make_grid_args(nx, ny):
    """
    Return arguments for a small grid.

    Parameters
    ----------
    nx : int
        The number of cells spanned by the longitude.
    ny : int
        The number of cells spanned by the latutude

    Returns
    -------
    Tuple
        Arguments which can be passed to
        :class:`~esmf_regrid.esmf_regridder.GridInfo.make_esmf_field`
    """
    small_grid_lon = np.array(range(nx)) * 10 / nx
    small_grid_lat = np.array(range(ny)) * 10 / ny

    small_grid_lon_bounds = np.array(range(nx + 1)) * 10 / nx
    small_grid_lat_bounds = np.array(range(ny + 1)) * 10 / ny
    return (
        small_grid_lon,
        small_grid_lat,
        small_grid_lon_bounds,
        small_grid_lat_bounds,
    )



"""Provides :mod:`esmpy` representations of UGRID meshes."""

import numpy as np

from .. import esmpy
from .._esmf_sdo import SDO


class MeshInfo(SDO):
    """
    Class for handling unstructured meshes.

    This class holds information about Meshes in a form similar to UGRID.
    It contains methods for translating this information into :mod:`esmpy` objects.
    In particular, there are methods for representing as an :class:`esmpy.api.mesh.Mesh`
    and as an :class:`esmpy.api.field.Field` containing that
    :class:`~esmpy.api.mesh.Mesh`. This esmpy :class:`~esmpy.api.field.Field` is designed to
    contain enough information for area weighted regridding and may be
    inappropriate for other :mod:`esmpy` regridding schemes.
    """

    def __init__(
        self,
        node_coords,
        face_node_connectivity,
        node_start_index,
        elem_start_index=0,
        areas=None,
        mask=None,
        elem_coords=None,
        location="face",
    ):
        """
        Create a :class:`MeshInfo` object describing a UGRID-like mesh.

        Parameters
        ----------
        node_coords: :obj:`~numpy.typing.ArrayLike`
            An ``Nx2`` array describing the location of the nodes of the mesh.
            ``node_coords[:,0]`` describes the longitudes in degrees and
            ``node_coords[:,1]`` describes the latitudes in degrees.
        face_node_connectivity: :obj:`~numpy.typing.ArrayLike`
            A masked array describing the face node connectivity of the
            mesh. The unmasked points of ``face_node_connectivity[i]`` describe
            which nodes are connected to the ``i``'th face.
        node_start_index: int
            A value which, appearing in the ``face_node_connectivity``
            array, indicates the first node in the ``node_coords`` array.
            UGRID supports both ``0`` based and ``1`` based indexing, so both must be
            accounted for here:
            https://ugrid-conventions.github.io/ugrid-conventions/#zero-or-one
        elem_start_index: int, default=0
            Describes what index should be considered by :mod:`esmpy` to be
            the start index for describing its elements. This makes no
            difference to the regridding calculation and will only affect the
            intermediate :mod:`esmpy` objects, should the user need access to them.
        areas: :obj:`~numpy.typing.ArrayLike`, optional
            Array describing the areas associated with
            each face. If ``None``, then :mod:`esmpy` will use its own calculated areas.
        mask: :obj:`~numpy.typing.ArrayLike`, optional
            Array describing which elements :mod:`esmpy` will ignore.
        elem_coords : :obj:`~numpy.typing.ArrayLike`, optional
            An ``Nx2`` array describing the location of the face centers of the mesh.
            ``elem_coords[:,0]`` describes the longitudes in degrees and
            ``elem_coords[:,1]`` describes the latitudes in degrees.
        location : str, default="face"
            Either "face" or "node". Describes the location for data on the mesh.
        """
        self.node_coords = node_coords
        self.fnc = face_node_connectivity
        self.nsi = node_start_index
        self.esi = elem_start_index
        self.areas = areas
        self.elem_coords = elem_coords
        if location == "face":
            field_kwargs = {"meshloc": esmpy.MeshLoc.ELEMENT}
            shape = (len(face_node_connectivity),)
        elif location == "node":
            field_kwargs = {"meshloc": esmpy.MeshLoc.NODE}
            shape = (len(node_coords),)
        else:
            raise ValueError(
                f"The mesh location '{location}' is not supported, only "
                f"'face' and 'node' are supported."
            )
        super().__init__(
            shape=shape,
            index_offset=1,
            field_kwargs=field_kwargs,
            mask=mask,
        )

    def _as_esmf_info(self):
        # ESMF uses a slightly different format to UGRID,
        # the data must be translated into a form ESMF understands
        num_node = self.node_coords.shape[0]
        num_elem = self.fnc.shape[0]
        nodeId = np.arange(1, num_node + 1)
        nodeCoord = self.node_coords.flatten()
        nodeOwner = np.zeros([num_node])  # regridding currently serial
        elemId = np.arange(1, num_elem + 1)
        elemType = self.fnc.count(axis=1)
        # Experiments seem to indicate that ESMF is using 0 indexing here
        elemConn = self.fnc.compressed() - self.nsi
        elemCoord = self.elem_coords
        result = (
            num_node,
            num_elem,
            nodeId,
            nodeCoord,
            nodeOwner,
            elemId,
            elemType,
            elemConn,
            self.mask,
            self.areas,
            elemCoord,
        )
        return result

    def _make_esmf_sdo(self):
        info = self._as_esmf_info()
        (
            num_node,
            num_elem,
            nodeId,
            nodeCoord,
            nodeOwner,
            elemId,
            elemType,
            elemConn,
            mask,
            areas,
            elemCoord,
        ) = info
        # ESMF can handle other dimensionalities, but we are unlikely
        # to make such a use of ESMF
        emesh = esmpy.Mesh(
            parametric_dim=2, spatial_dim=2, coord_sys=esmpy.CoordSys.SPH_DEG
        )

        emesh.add_nodes(num_node, nodeId, nodeCoord, nodeOwner)
        emesh.add_elements(
            num_elem,
            elemId,
            elemType,
            elemConn,
            element_mask=mask,
            element_area=areas,
            element_coords=elemCoord,
        )
        return emesh



"""Unit tests for :func:`esmf_regrid.schemes.regrid_rectilinear_to_rectilinear`."""

import dask.array as da
from iris.coord_systems import RotatedGeogCS
from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube
import numpy as np
from numpy import ma
import pytest

from esmf_regrid.schemes import regrid_rectilinear_to_rectilinear
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _curvilinear_cube,
    _grid_cube,
)


def test_rotated_regridding():
    """
    Test for :func:`esmf_regrid.schemes.regrid_rectilinear_to_rectilinear`.

    Test the regriding of a rotated pole coordinate system. The test is
    designed to that it should be possible to verify the result by
    inspection.
    """
    src_coord_system = RotatedGeogCS(0, 90, 90)
    tgt_coord_system = None

    n_lons = 4
    n_lats = 4
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)

    src = _grid_cube(
        n_lons,
        n_lats,
        lon_bounds,
        lat_bounds,
        circular=True,
        coord_system=src_coord_system,
    )
    tgt = _grid_cube(
        n_lons,
        n_lats,
        lon_bounds,
        lat_bounds,
        circular=True,
        coord_system=tgt_coord_system,
    )
    src_data = np.arange(n_lons * n_lats).reshape([n_lats, n_lons])
    # src_mask = np.empty([n_lats, n_lons])
    # src_mask[:] = np.array([1, 0, 0, 1])[:, np.newaxis]
    src_mask = np.array([[1, 1, 1, 1], [0, 0, 0, 0], [0, 0, 0, 0], [1, 1, 1, 1]])
    src_data = ma.array(src_data, mask=src_mask)
    src.data = src_data

    no_mdtol_result = regrid_rectilinear_to_rectilinear(src, tgt)

    full_mdtol_result = regrid_rectilinear_to_rectilinear(src, tgt, mdtol=1)

    expected_data = np.array(
        [[5, 4, 8, 9], [5, 4, 8, 9], [6, 7, 11, 10], [6, 7, 11, 10]]
    )
    expected_mask = np.array([[0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 0, 0]])
    no_mdtol_expected_data = ma.array(expected_data, mask=expected_mask)

    # Lenient check for data.
    assert np.allclose(no_mdtol_expected_data, no_mdtol_result.data)
    assert np.allclose(expected_data, full_mdtol_result.data)


def test_extra_dims():
    """
    Test for :func:`esmf_regrid.schemes.regrid_rectilinear_to_rectilinear`.

    Tests the handling of extra dimensions and metadata. Ensures that proper
    coordinates, attributes, names and units are copied over.
    """
    h = 2
    t = 4
    e = 6
    src_lats = 3
    src_lons = 5

    tgt_lats = 5
    tgt_lons = 3

    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)

    src_grid = _grid_cube(
        src_lons,
        src_lats,
        lon_bounds,
        lat_bounds,
    )
    tgt_grid = _grid_cube(
        tgt_lons,
        tgt_lats,
        lon_bounds,
        lat_bounds,
    )

    height = DimCoord(np.arange(h), standard_name="height")
    time = DimCoord(np.arange(t), standard_name="time")
    extra = AuxCoord(np.arange(e), long_name="extra dim")
    spanning = AuxCoord(np.ones([h, t, e]), long_name="spanning dim")

    src_data = np.empty([h, src_lats, t, src_lons, e])
    src_data[:] = np.arange(t * h * e).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]

    src_cube = Cube(src_data)
    src_cube.add_dim_coord(height, 0)
    src_cube.add_dim_coord(src_grid.coord("latitude"), 1)
    src_cube.add_dim_coord(time, 2)
    src_cube.add_dim_coord(src_grid.coord("longitude"), 3)
    src_cube.add_aux_coord(extra, 4)
    src_cube.add_aux_coord(spanning, [0, 2, 4])

    def _add_metadata(cube):
        result = cube.copy()
        result.units = "K"
        result.attributes = {"a": 1}
        result.standard_name = "air_temperature"
        scalar_height = AuxCoord([5], units="m", standard_name="height")
        scalar_time = DimCoord([10], units="s", standard_name="time")
        result.add_aux_coord(scalar_height)
        result.add_aux_coord(scalar_time)
        return result

    src_cube = _add_metadata(src_cube)

    result = regrid_rectilinear_to_rectilinear(src_cube, tgt_grid)

    expected_data = np.empty([h, tgt_lats, t, tgt_lons, e])
    expected_data[:] = np.arange(t * h * e).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]

    expected_cube = Cube(expected_data)
    expected_cube.add_dim_coord(height, 0)
    expected_cube.add_dim_coord(tgt_grid.coord("latitude"), 1)
    expected_cube.add_dim_coord(time, 2)
    expected_cube.add_dim_coord(tgt_grid.coord("longitude"), 3)
    expected_cube.add_aux_coord(extra, 4)
    expected_cube.add_aux_coord(spanning, [0, 2, 4])
    expected_cube = _add_metadata(expected_cube)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)

    # Check metadata and coords.
    result.data = expected_data
    assert expected_cube == result


@pytest.mark.parametrize("src_transposed", (False, True))
@pytest.mark.parametrize("tgt_transposed", (False, True))
def test_laziness(src_transposed, tgt_transposed):
    """Test that regridding is lazy when source data is lazy."""
    n_lons = 12
    n_lats = 10
    h = 4
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)

    grid = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    src_data = np.arange(n_lats * n_lons * h, dtype=np.float32).reshape(
        [n_lats, n_lons, h]
    )
    src_data = da.from_array(src_data, chunks=[3, 5, 1])
    src = Cube(src_data)
    src.add_dim_coord(grid.coord("latitude"), 0)
    src.add_dim_coord(grid.coord("longitude"), 1)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    # Transpose the data to ensure all possible combinations of dimension
    # ordering are covered.
    if src_transposed:
        src.transpose()
        src_data = src_data.T
    if tgt_transposed:
        tgt.transpose()

    assert src.has_lazy_data()
    result = regrid_rectilinear_to_rectilinear(src, tgt)
    assert result.has_lazy_data()
    assert result.lazy_data().dtype == np.float64
    assert result.data.dtype == np.float64
    assert np.allclose(result.data, src_data)


@pytest.mark.parametrize("src_transposed", (False, True))
@pytest.mark.parametrize("tgt_transposed", (False, True))
def test_laziness_curvilinear(src_transposed, tgt_transposed):
    """
    Test for :func:`esmf_regrid.schemes.regrid_rectilinear_to_rectilinear`.

    Tests the handling of lazy data. Ensures that handling is the same
    for non-lazy data.
    """
    h = 2
    t = 4
    e = 6
    src_lats = 3
    src_lons = 5

    tgt_lats = 5
    tgt_lons = 3

    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)

    src_grid = _curvilinear_cube(
        src_lons,
        src_lats,
        lon_bounds,
        lat_bounds,
    )
    tgt_grid = _curvilinear_cube(
        tgt_lons,
        tgt_lats,
        lon_bounds,
        lat_bounds,
    )

    height = DimCoord(np.arange(h), standard_name="height")
    time = DimCoord(np.arange(t), standard_name="time")
    extra = AuxCoord(np.arange(e), long_name="extra dim")

    src_data = np.empty([h, src_lats, t, src_lons, e])
    src_data[:] = np.arange(t * h * e, dtype=np.float32).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]
    src_data_lazy = da.array(src_data)

    src_cube = Cube(src_data)
    src_cube.add_dim_coord(height, 0)
    src_cube.add_aux_coord(src_grid.coord("latitude"), (1, 3))
    src_cube.add_dim_coord(time, 2)
    src_cube.add_aux_coord(src_grid.coord("longitude"), (1, 3))
    src_cube.add_aux_coord(extra, 4)

    src_cube_lazy = src_cube.copy(src_data_lazy)

    # Transpose the data to ensure all possible combinations of dimension
    # ordering are covered.
    if src_transposed:
        src_cube_lazy.transpose()
        src_cube.transpose()
    if tgt_transposed:
        tgt_grid.transpose()

    result = regrid_rectilinear_to_rectilinear(src_cube, tgt_grid)
    result_lazy = regrid_rectilinear_to_rectilinear(src_cube_lazy, tgt_grid)

    assert result_lazy.has_lazy_data()
    assert result.lazy_data().dtype == np.float64
    assert result.data.dtype == np.float64

    assert result_lazy == result


def test_extra_dims_curvilinear():
    """
    Test for :func:`esmf_regrid.schemes.regrid_rectilinear_to_rectilinear`.

    Tests the handling of extra dimensions and metadata. Ensures that proper
    coordinates, attributes, names and units are copied over.
    """
    h = 2
    t = 4
    e = 6
    src_lats = 3
    src_lons = 5

    tgt_lats = 5
    tgt_lons = 3

    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)

    src_grid = _curvilinear_cube(
        src_lons,
        src_lats,
        lon_bounds,
        lat_bounds,
    )
    tgt_grid = _curvilinear_cube(
        tgt_lons,
        tgt_lats,
        lon_bounds,
        lat_bounds,
    )

    height = DimCoord(np.arange(h), standard_name="height")
    time = DimCoord(np.arange(t), standard_name="time")
    extra = AuxCoord(np.arange(e), long_name="extra dim")
    spanning = AuxCoord(np.ones([h, t, e]), long_name="spanning dim")

    src_data = np.empty([h, src_lats, t, src_lons, e])
    src_data[:] = np.arange(t * h * e).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]

    src_cube = Cube(src_data)
    src_cube.add_dim_coord(height, 0)
    src_cube.add_aux_coord(src_grid.coord("latitude"), (1, 3))
    src_cube.add_dim_coord(time, 2)
    src_cube.add_aux_coord(src_grid.coord("longitude"), (1, 3))
    src_cube.add_aux_coord(extra, 4)
    src_cube.add_aux_coord(spanning, [0, 2, 4])

    def _add_metadata(cube):
        result = cube.copy()
        result.units = "K"
        result.attributes = {"a": 1}
        result.standard_name = "air_temperature"
        scalar_height = AuxCoord([5], units="m", standard_name="height")
        scalar_time = DimCoord([10], units="s", standard_name="time")
        result.add_aux_coord(scalar_height)
        result.add_aux_coord(scalar_time)
        return result

    src_cube = _add_metadata(src_cube)

    result = regrid_rectilinear_to_rectilinear(src_cube, tgt_grid)

    expected_data = np.empty([h, tgt_lats, t, tgt_lons, e])
    expected_data[:] = np.arange(t * h * e).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]

    expected_cube = Cube(expected_data)
    expected_cube.add_dim_coord(height, 0)
    expected_cube.add_aux_coord(tgt_grid.coord("latitude"), (1, 3))
    expected_cube.add_dim_coord(time, 2)
    expected_cube.add_aux_coord(tgt_grid.coord("longitude"), (1, 3))
    expected_cube.add_aux_coord(extra, 4)
    expected_cube.add_aux_coord(spanning, [0, 2, 4])
    expected_cube = _add_metadata(expected_cube)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)

    # Check metadata and coords.
    result.data = expected_data
    assert expected_cube == result


def test_extra_dims_curvilinear_to_rectilinear():
    """
    Test for :func:`esmf_regrid.schemes.regrid_rectilinear_to_rectilinear`.

    Tests the handling of extra dimensions and metadata. Ensures that proper
    coordinates, attributes, names and units are copied over.
    """
    h = 2
    t = 4
    e = 6
    src_lats = 3
    src_lons = 5

    tgt_lats = 5
    tgt_lons = 3

    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)

    src_grid = _curvilinear_cube(
        src_lons,
        src_lats,
        lon_bounds,
        lat_bounds,
    )
    tgt_grid = _grid_cube(
        tgt_lons,
        tgt_lats,
        lon_bounds,
        lat_bounds,
    )

    height = DimCoord(np.arange(h), standard_name="height")
    time = DimCoord(np.arange(t), standard_name="time")
    extra = AuxCoord(np.arange(e), long_name="extra dim")
    spanning = AuxCoord(np.ones([h, t, e]), long_name="spanning dim")

    src_data = np.empty([h, src_lats, t, src_lons, e])
    src_data[:] = np.arange(t * h * e).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]

    src_cube = Cube(src_data)
    src_cube.add_dim_coord(height, 0)
    src_cube.add_aux_coord(src_grid.coord("latitude"), (1, 3))
    src_cube.add_dim_coord(time, 2)
    src_cube.add_aux_coord(src_grid.coord("longitude"), (1, 3))
    src_cube.add_aux_coord(extra, 4)
    src_cube.add_aux_coord(spanning, [0, 2, 4])

    def _add_metadata(cube):
        result = cube.copy()
        result.units = "K"
        result.attributes = {"a": 1}
        result.standard_name = "air_temperature"
        scalar_height = AuxCoord([5], units="m", standard_name="height")
        scalar_time = DimCoord([10], units="s", standard_name="time")
        result.add_aux_coord(scalar_height)
        result.add_aux_coord(scalar_time)
        return result

    src_cube = _add_metadata(src_cube)

    result = regrid_rectilinear_to_rectilinear(src_cube, tgt_grid)

    expected_data = np.empty([h, tgt_lats, t, tgt_lons, e])
    expected_data[:] = np.arange(t * h * e).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]

    expected_cube = Cube(expected_data)
    expected_cube.add_dim_coord(height, 0)
    expected_cube.add_dim_coord(tgt_grid.coord("latitude"), 1)
    expected_cube.add_dim_coord(time, 2)
    expected_cube.add_dim_coord(tgt_grid.coord("longitude"), 3)
    expected_cube.add_aux_coord(extra, 4)
    expected_cube.add_aux_coord(spanning, [0, 2, 4])
    expected_cube = _add_metadata(expected_cube)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)

    # Check metadata and coords.
    result.data = expected_data
    assert expected_cube == result


def test_extra_dims_rectilinear_to_curvilinear():
    """
    Test for :func:`esmf_regrid.schemes.regrid_rectilinear_to_rectilinear`.

    Tests the handling of extra dimensions and metadata. Ensures that proper
    coordinates, attributes, names and units are copied over.
    """
    h = 2
    t = 4
    e = 6
    src_lats = 3
    src_lons = 5

    tgt_lats = 5
    tgt_lons = 3

    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)

    src_grid = _grid_cube(
        src_lons,
        src_lats,
        lon_bounds,
        lat_bounds,
    )
    tgt_grid = _curvilinear_cube(
        tgt_lons,
        tgt_lats,
        lon_bounds,
        lat_bounds,
    )

    height = DimCoord(np.arange(h), standard_name="height")
    time = DimCoord(np.arange(t), standard_name="time")
    extra = AuxCoord(np.arange(e), long_name="extra dim")
    spanning = AuxCoord(np.ones([h, t, e]), long_name="spanning dim")

    src_data = np.empty([h, src_lats, t, src_lons, e])
    src_data[:] = np.arange(t * h * e).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]

    src_cube = Cube(src_data)
    src_cube.add_dim_coord(height, 0)
    src_cube.add_dim_coord(src_grid.coord("latitude"), 1)
    src_cube.add_dim_coord(time, 2)
    src_cube.add_dim_coord(src_grid.coord("longitude"), 3)
    src_cube.add_aux_coord(extra, 4)
    src_cube.add_aux_coord(spanning, [0, 2, 4])

    def _add_metadata(cube):
        result = cube.copy()
        result.units = "K"
        result.attributes = {"a": 1}
        result.standard_name = "air_temperature"
        scalar_height = AuxCoord([5], units="m", standard_name="height")
        scalar_time = DimCoord([10], units="s", standard_name="time")
        result.add_aux_coord(scalar_height)
        result.add_aux_coord(scalar_time)
        return result

    src_cube = _add_metadata(src_cube)

    result = regrid_rectilinear_to_rectilinear(src_cube, tgt_grid)

    expected_data = np.empty([h, tgt_lats, t, tgt_lons, e])
    expected_data[:] = np.arange(t * h * e).reshape([h, t, e])[
        :, np.newaxis, :, np.newaxis, :
    ]

    expected_cube = Cube(expected_data)
    expected_cube.add_dim_coord(height, 0)
    expected_cube.add_aux_coord(tgt_grid.coord("latitude"), (1, 3))
    expected_cube.add_dim_coord(time, 2)
    expected_cube.add_aux_coord(tgt_grid.coord("longitude"), (1, 3))
    expected_cube.add_aux_coord(extra, 4)
    expected_cube.add_aux_coord(spanning, [0, 2, 4])
    expected_cube = _add_metadata(expected_cube)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)

    # Check metadata and coords.
    result.data = expected_data
    assert expected_cube == result



"""Unit tests for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`."""

from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube
import numpy as np
from numpy import ma
import pytest

from esmf_regrid import Constants
from esmf_regrid.experimental.unstructured_scheme import (
    regrid_rectilinear_to_unstructured,
)
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _grid_cube,
)
from esmf_regrid.tests.unit.schemes.test__mesh_to_MeshInfo import (
    _gridlike_mesh_cube,
)
from esmf_regrid.tests.unit.schemes.test__regrid_unstructured_to_rectilinear__prepare import (
    _flat_mesh_cube,
)


def _add_metadata(cube):
    result = cube.copy()
    result.units = "K"
    result.attributes = {"a": 1}
    result.standard_name = "air_temperature"
    scalar_height = AuxCoord([5], units="m", standard_name="height")
    scalar_time = DimCoord([10], units="s", standard_name="time")
    result.add_aux_coord(scalar_height)
    result.add_aux_coord(scalar_time)
    return result


def test_flat_cubes():
    """
    Basic test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`.

    Tests with flat cubes as input (a 2D grid cube and a 1D mesh cube).
    """
    tgt = _flat_mesh_cube()

    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    # Ensure data in the target grid is different to the expected data.
    # i.e. target grid data is all zero, expected data is all one
    tgt.data[:] = 0

    src = _add_metadata(src)
    src.data[:] = 1  # Ensure all data in the source is one.
    result = regrid_rectilinear_to_unstructured(src, tgt)
    src_T = src.copy()
    src_T.transpose()
    result_transposed = regrid_rectilinear_to_unstructured(src_T, tgt)

    expected_data = np.ones_like(tgt.data)
    expected_cube = _add_metadata(tgt)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)
    assert np.allclose(expected_data, result_transposed.data)

    # Check metadata and scalar coords.
    expected_cube.data = result.data
    assert expected_cube == result
    expected_cube.data = result_transposed.data
    assert expected_cube == result_transposed


@pytest.mark.parametrize(
    "method", (Constants.Method.BILINEAR, Constants.Method.NEAREST)
)
def test_node_friendly_methods(method):
    """
    Basic test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`.

    Tests with the bilinear and nearest method.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    tgt = _gridlike_mesh_cube(n_lons, n_lats, location="node")
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    # Ensure data in the target grid is different to the expected data.
    # i.e. target grid data is all zero, expected data is all one
    tgt.data[:] = 0

    src = _add_metadata(src)
    src.data[:] = 1  # Ensure all data in the source is one.
    result = regrid_rectilinear_to_unstructured(src, tgt, method=method)

    expected_data = np.ones_like(tgt.data)
    expected_cube = _add_metadata(tgt)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)

    # Check metadata and scalar coords.
    expected_cube.data = result.data
    assert expected_cube == result


def test_invalid_args():
    """
    Test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`.

    Tests that an appropriate error is raised when arguments are invalid.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    node_tgt = _gridlike_mesh_cube(n_lons, n_lats, location="node")
    edge_tgt = _gridlike_mesh_cube(n_lons, n_lats, location="edge")
    face_tgt = _gridlike_mesh_cube(n_lons, n_lats, location="face")
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    with pytest.raises(ValueError):
        _ = regrid_rectilinear_to_unstructured(
            src, src, method=Constants.Method.BILINEAR
        )
    with pytest.raises(ValueError):
        _ = regrid_rectilinear_to_unstructured(src, face_tgt, method="other")
    with pytest.raises(ValueError) as excinfo:
        _ = regrid_rectilinear_to_unstructured(
            src, node_tgt, method=Constants.Method.CONSERVATIVE
        )
    expected_message = (
        "conservative regridding requires a target cube located on "
        "the face of a cube, target cube had the node location."
    )
    assert expected_message in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        _ = regrid_rectilinear_to_unstructured(
            src, edge_tgt, method=Constants.Method.BILINEAR
        )
    expected_message = (
        "bilinear regridding requires a target cube with a node "
        "or face location, target cube had the edge location."
    )
    assert expected_message in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        _ = regrid_rectilinear_to_unstructured(
            src, edge_tgt, method=Constants.Method.NEAREST
        )
    expected_message = (
        "nearest regridding requires a target cube with a node "
        "or face location, target cube had the edge location."
    )
    assert expected_message in str(excinfo.value)


def test_multidim_cubes():
    """
    Test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`.

    Tests with multidimensional cubes. The source cube contains
    coordinates on the dimensions before and after the grid dimensions.
    """
    tgt = _flat_mesh_cube()
    mesh = tgt.mesh
    mesh_length = mesh.connectivity(contains_face=True).shape[0]
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    h = 2
    p = 4
    t = 3
    height = DimCoord(np.arange(h), standard_name="height")
    pressure = DimCoord(np.arange(p), standard_name="air_pressure")
    time = DimCoord(np.arange(t), standard_name="time")
    spanning = AuxCoord(np.ones([t, p, h]), long_name="spanning dim")
    ignore = AuxCoord(np.ones([n_lats, h]), long_name="ignore")

    src_data = np.empty([t, n_lats, p, n_lons, h])
    src_data[:] = np.arange(t * p * h).reshape([t, p, h])[
        :, np.newaxis, :, np.newaxis, :
    ]
    cube = Cube(src_data)
    cube.add_dim_coord(grid.coord("latitude"), 1)
    cube.add_dim_coord(grid.coord("longitude"), 3)
    cube.add_dim_coord(time, 0)
    cube.add_dim_coord(pressure, 2)
    cube.add_dim_coord(height, 4)
    cube.add_aux_coord(spanning, [0, 2, 4])
    cube.add_aux_coord(ignore, [1, 4])

    result = regrid_rectilinear_to_unstructured(cube, tgt)

    cube_transposed = cube.copy()
    cube_transposed.transpose([0, 3, 2, 1, 4])
    result_transposed = regrid_rectilinear_to_unstructured(cube_transposed, tgt)

    # Lenient check for data.
    expected_data = np.empty([t, mesh_length, p, h])
    expected_data[:] = np.arange(t * p * h).reshape(t, p, h)[:, np.newaxis, :, :]
    assert np.allclose(expected_data, result.data)
    assert np.allclose(expected_data, result_transposed.data)

    mesh_coord_x, mesh_coord_y = mesh.to_MeshCoords("face")
    expected_cube = Cube(expected_data)
    expected_cube.add_dim_coord(time, 0)
    expected_cube.add_aux_coord(mesh_coord_x, 1)
    expected_cube.add_aux_coord(mesh_coord_y, 1)
    expected_cube.add_dim_coord(pressure, 2)
    expected_cube.add_dim_coord(height, 3)
    expected_cube.add_aux_coord(spanning, [0, 2, 3])

    # Check metadata and scalar coords.
    result.data = expected_data
    assert expected_cube == result
    result_transposed.data = expected_data
    assert expected_cube == result_transposed


def test_mask_handling():
    """
    Test masked data handling for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`.

    Tests masked data handling for multiple valid values for mdtol.
    """
    tgt = _flat_mesh_cube()

    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    data = np.ones([n_lats, n_lons])
    mask = np.zeros([n_lats, n_lons])
    mask[0, 0] = 1
    masked_data = ma.array(data, mask=mask)
    src.data = masked_data
    result_0 = regrid_rectilinear_to_unstructured(src, tgt, mdtol=0)
    result_05 = regrid_rectilinear_to_unstructured(src, tgt, mdtol=0.05)
    result_1 = regrid_rectilinear_to_unstructured(src, tgt, mdtol=1)

    expected_data = np.ones(tgt.shape)
    expected_0 = ma.array(expected_data)
    expected_05 = ma.array(expected_data, mask=[0, 0, 1, 0, 0, 0])
    expected_1 = ma.array(expected_data, mask=[1, 0, 1, 0, 0, 0])

    assert ma.allclose(expected_0, result_0.data)
    assert ma.allclose(expected_05, result_05.data)
    assert ma.allclose(expected_1, result_1.data)


def test_resolution():
    """
    Basic test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`.

    Tests the src_resolution keyword with grids that would otherwise not work.
    """
    tgt = _flat_mesh_cube()

    # The resulting grid has full latitude bounds and cells must be split up.
    n_lons = 1
    n_lats = 4
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds)
    # Ensure data in the target grid is different to the expected data.
    # i.e. target grid data is all zero, expected data is all one
    tgt.data[:] = 0

    src = _add_metadata(src)
    src.data[:] = 1  # Ensure all data in the source is one.
    result = regrid_rectilinear_to_unstructured(src, tgt, src_resolution=8)

    expected_data = np.ones_like(tgt.data)
    expected_cube = _add_metadata(tgt)

    # Lenient check for data.
    # Note that when src_resolution=None, this would be a fully masked array.
    assert np.allclose(expected_data, result.data)

    # Check metadata and scalar coords.
    expected_cube.data = result.data
    assert expected_cube == result



"""Holds all enums created for esmf-regrid."""

from enum import Enum

from . import esmpy


class Constants:
    """Encompassing class for best practice import."""

    class Method(Enum):
        """holds enums for Method values."""

        CONSERVATIVE = esmpy.RegridMethod.CONSERVE
        BILINEAR = esmpy.RegridMethod.BILINEAR
        NEAREST = esmpy.RegridMethod.NEAREST_STOD

    class NormType(Enum):
        """holds enums for norm types."""

        FRACAREA = esmpy.api.constants.NormType.FRACAREA
        DSTAREA = esmpy.api.constants.NormType.DSTAREA


method_dict = {
    "conservative": Constants.Method.CONSERVATIVE,
    "bilinear": Constants.Method.BILINEAR,
    "nearest": Constants.Method.NEAREST,
}

norm_dict = {
    "fracarea": Constants.NormType.FRACAREA,
    "dstarea": Constants.NormType.DSTAREA,
}


def check_method(method):
    """Check that method is a member of the `Constants.Method` enum or raise an error."""
    if method in method_dict.keys():
        result = method_dict[method]
    elif method in method_dict.values():
        result = method
    else:
        raise ValueError(
            f"Method must be a member of `Constants.Method` enum, instead got {method}"
        )
    return result


def check_norm(norm):
    """Check that normtype is a member of the `Constants.NormType` enum or raise an error."""
    if norm in norm_dict.keys():
        result = norm_dict[norm]
    elif norm in norm_dict.values():
        result = norm
    else:
        raise ValueError(
            f"NormType must be a member of `Constants.NormType` enum, instead got {norm}"
        )
    return result



"""Unit tests for :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`."""

import dask.array as da
from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube
import numpy as np
from numpy import ma
import pytest

from esmf_regrid import Constants
from esmf_regrid.experimental.unstructured_scheme import (
    GridToMeshESMFRegridder,
)
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _curvilinear_cube,
    _grid_cube,
)
from esmf_regrid.tests.unit.schemes.test__mesh_to_MeshInfo import (
    _gridlike_mesh,
    _gridlike_mesh_cube,
)
from esmf_regrid.tests.unit.schemes.test__regrid_unstructured_to_rectilinear__prepare import (
    _flat_mesh_cube,
)


def _add_metadata(cube):
    result = cube.copy()
    result.units = "K"
    result.attributes = {"a": 1}
    result.standard_name = "air_temperature"
    scalar_height = AuxCoord([5], units="m", standard_name="height")
    scalar_time = DimCoord([10], units="s", standard_name="time")
    result.add_aux_coord(scalar_height)
    result.add_aux_coord(scalar_time)
    return result


def test_flat_cubes():
    """
    Basic test for :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Tests with flat cubes as input (a 2D grid cube and a 1D mesh cube).
    """
    tgt = _flat_mesh_cube()

    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    # Ensure data in the target grid is different to the expected data.
    # i.e. target grid data is all zero, expected data is all one
    tgt.data[:] = 0

    src = _add_metadata(src)
    src.data[:] = 1  # Ensure all data in the source is one.
    regridder = GridToMeshESMFRegridder(src, tgt)
    result = regridder(src)
    src_T = src.copy()
    src_T.transpose()
    result_transposed = regridder(src_T)

    expected_data = np.ones([n_lats, n_lons])
    expected_cube = _add_metadata(tgt)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)
    assert np.allclose(expected_data, result_transposed.data)

    # Check metadata and scalar coords.
    expected_cube.data = result.data
    assert expected_cube == result
    expected_cube.data = result_transposed.data
    assert expected_cube == result_transposed


@pytest.mark.parametrize("nsi", [0, 1])
@pytest.mark.parametrize(
    "method", [Constants.Method.BILINEAR, Constants.Method.NEAREST]
)
def test_node_friendly_methods(method, nsi):
    """
    Basic test for :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Tests with method=Constants.Method.BILINEAR and method=Constants.Method.NEAREST.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    face_tgt = _gridlike_mesh_cube(n_lons, n_lats, location="face", nsi=nsi)
    node_tgt = _gridlike_mesh_cube(n_lons, n_lats, location="node", nsi=nsi)

    src = _add_metadata(src)
    src.data[:] = 1  # Ensure all data in the source is one.
    face_regridder = GridToMeshESMFRegridder(src, face_tgt, method=method)
    node_regridder = GridToMeshESMFRegridder(src, node_tgt, method=method)

    assert face_regridder.regridder.method == method
    assert node_regridder.regridder.method == method

    face_expected_data = np.ones_like(face_tgt.data)
    node_expected_data = np.ones_like(node_tgt.data)
    face_result = face_regridder(src)
    node_result = node_regridder(src)

    # Lenient check for data.
    assert np.allclose(face_expected_data, face_result.data)
    assert np.allclose(node_expected_data, node_result.data)

    # Check metadata and scalar coords.
    face_expected_cube = _add_metadata(face_tgt)
    node_expected_cube = _add_metadata(node_tgt)
    face_expected_cube.data = face_result.data
    node_expected_cube.data = node_result.data
    assert face_expected_cube == face_result
    assert node_expected_cube == node_result


def test_multidim_cubes():
    """
    Test for :class:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`.

    Tests with multidimensional cubes. The source cube contains
    coordinates on the dimensions before and after the grid dimensions.
    """
    tgt = _flat_mesh_cube()
    mesh = tgt.mesh
    mesh_length = mesh.connectivity(contains_face=True).shape[0]
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    h = 2
    t = 3
    height = DimCoord(np.arange(h), standard_name="height")
    time = DimCoord(np.arange(t), standard_name="time")

    src_data = np.empty([t, n_lats, n_lons, h])
    src_data[:] = np.arange(t * h).reshape([t, h])[:, np.newaxis, np.newaxis, :]
    cube = Cube(src_data)
    cube.add_dim_coord(grid.coord("latitude"), 1)
    cube.add_dim_coord(grid.coord("longitude"), 2)
    cube.add_dim_coord(time, 0)
    cube.add_dim_coord(height, 3)

    regridder = GridToMeshESMFRegridder(grid, tgt)
    result = regridder(cube)

    # Lenient check for data.
    expected_data = np.empty([t, mesh_length, h])
    expected_data[:] = np.arange(t * h).reshape(t, h)[:, np.newaxis, :]
    assert np.allclose(expected_data, result.data)

    mesh_coord_x, mesh_coord_y = mesh.to_MeshCoords("face")
    expected_cube = Cube(expected_data)
    expected_cube.add_dim_coord(time, 0)
    expected_cube.add_aux_coord(mesh_coord_x, 1)
    expected_cube.add_aux_coord(mesh_coord_y, 1)
    expected_cube.add_dim_coord(height, 2)

    # Check metadata and scalar coords.
    result.data = expected_data
    assert expected_cube == result


def test_invalid_mdtol():
    """
    Test initialisation of :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Checks that an error is raised when mdtol is out of range.
    """
    tgt = _flat_mesh_cube()

    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    with pytest.raises(ValueError):
        _ = GridToMeshESMFRegridder(src, tgt, mdtol=2)
    with pytest.raises(ValueError):
        _ = GridToMeshESMFRegridder(src, tgt, mdtol=-1)


def test_invalid_method():
    """
    Test initialisation of :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Checks that an error is raised when the method is invalid.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    face_tgt = _gridlike_mesh_cube(n_lons, n_lats, location="face")
    edge_tgt = _gridlike_mesh_cube(n_lons, n_lats, location="edge")
    node_tgt = _gridlike_mesh_cube(n_lons, n_lats, location="node")
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    with pytest.raises(ValueError):
        _ = GridToMeshESMFRegridder(src, face_tgt, method="other")
    with pytest.raises(ValueError) as excinfo:
        _ = GridToMeshESMFRegridder(src, node_tgt, method=Constants.Method.CONSERVATIVE)
    expected_message = (
        "conservative regridding requires a target cube located on "
        "the face of a cube, target cube had the node location."
    )
    assert expected_message in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        _ = GridToMeshESMFRegridder(src, edge_tgt, method=Constants.Method.BILINEAR)
    expected_message = (
        "bilinear regridding requires a target cube with a node "
        "or face location, target cube had the edge location."
    )
    assert expected_message in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        _ = GridToMeshESMFRegridder(src, edge_tgt, method=Constants.Method.NEAREST)
    expected_message = (
        "nearest regridding requires a target cube with a node "
        "or face location, target cube had the edge location."
    )
    assert expected_message in str(excinfo.value)


def test_invalid_resolution():
    """
    Test initialisation of :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Checks that an error is raised when the resolution is invalid.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    tgt = _gridlike_mesh_cube(n_lons, n_lats, location="face")
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    with pytest.raises(ValueError) as excinfo:
        _ = GridToMeshESMFRegridder(
            src, tgt, method=Constants.Method.CONSERVATIVE, src_resolution=-1
        )
    expected_message = "resolution must be a positive integer."
    assert expected_message in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        _ = GridToMeshESMFRegridder(
            src, tgt, method=Constants.Method.BILINEAR, src_resolution=4
        )
    expected_message = "resolution can only be set for conservative regridding."
    assert expected_message in str(excinfo.value)


def test_default_mdtol():
    """
    Test initialisation of :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Checks that default mdtol values are as expected.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    tgt = _gridlike_mesh_cube(n_lons, n_lats)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    rg_con = GridToMeshESMFRegridder(src, tgt, method=Constants.Method.CONSERVATIVE)
    assert rg_con.mdtol == 1
    rg_bi = GridToMeshESMFRegridder(src, tgt, method=Constants.Method.BILINEAR)
    assert rg_bi.mdtol == 0


def test_mismatched_grids():
    """
    Test error handling in calling of :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Checks that an error is raised when the regridder is called with a
    cube whose grid does not match with the one used when initialising
    the regridder.
    """
    tgt = _flat_mesh_cube()
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    regridder = GridToMeshESMFRegridder(src, tgt)

    n_lons_other = 3
    n_lats_other = 10
    src_other = _grid_cube(
        n_lons_other, n_lats_other, lon_bounds, lat_bounds, circular=True
    )
    with pytest.raises(ValueError):
        _ = regridder(src_other)


def test_mask_handling():
    """
    Test masked data handling for :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Tests masked data handling for multiple valid values for mdtol.
    """
    tgt = _flat_mesh_cube()

    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    data = np.ones([n_lats, n_lons])
    mask = np.zeros([n_lats, n_lons])
    mask[0, 0] = 1
    masked_data = ma.array(data, mask=mask)
    src.data = masked_data
    regridder_0 = GridToMeshESMFRegridder(src, tgt, mdtol=0)
    regridder_05 = GridToMeshESMFRegridder(src, tgt, mdtol=0.05)
    regridder_1 = GridToMeshESMFRegridder(src, tgt, mdtol=1)
    result_0 = regridder_0(src)
    result_05 = regridder_05(src)
    result_1 = regridder_1(src)

    expected_data = np.ones(tgt.shape)
    expected_0 = ma.array(expected_data)
    expected_05 = ma.array(expected_data, mask=[0, 0, 1, 0, 0, 0])
    expected_1 = ma.array(expected_data, mask=[1, 0, 1, 0, 0, 0])

    assert ma.allclose(expected_0, result_0.data)
    assert ma.allclose(expected_05, result_05.data)
    assert ma.allclose(expected_1, result_1.data)


def test_laziness():
    """Test that regridding is lazy when source data is lazy."""
    n_lons = 12
    n_lats = 10
    h = 4
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)

    mesh = _gridlike_mesh(n_lons, n_lats)

    src_data = np.arange(n_lats * n_lons * h).reshape([n_lats, n_lons, h])
    src_data = da.from_array(src_data, chunks=[3, 5, 2])
    src = Cube(src_data)
    grid = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    src.add_dim_coord(grid.coord("latitude"), 0)
    src.add_dim_coord(grid.coord("longitude"), 1)

    mesh_coord_x, mesh_coord_y = mesh.to_MeshCoords("face")
    tgt_data = np.zeros([n_lats * n_lons])
    tgt = Cube(tgt_data)
    tgt.add_aux_coord(mesh_coord_x, 0)
    tgt.add_aux_coord(mesh_coord_y, 0)

    rg = GridToMeshESMFRegridder(src, tgt)

    assert src.has_lazy_data()
    result = rg(src)
    assert result.has_lazy_data()
    out_chunks = result.lazy_data().chunks
    expected_chunks = ((120,), (2, 2))
    assert out_chunks == expected_chunks
    assert np.allclose(result.data, src_data.reshape([-1, h]))


def test_resolution():
    """
    Test for :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Tests for the src_resolution keyword.
    """
    tgt = _flat_mesh_cube()
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    resolution = 8

    result = GridToMeshESMFRegridder(grid, tgt, src_resolution=resolution)
    assert result.resolution == resolution
    assert result.regridder.src.resolution == resolution


def test_curvilinear():
    """
    Test for :class:`esmf_regrid.experimental.unstructured_scheme.regrid_rectilinear_to_unstructured`.

    Tests with curvilinear target cube.
    """
    tgt = _flat_mesh_cube()
    mesh = tgt.mesh
    mesh_length = mesh.connectivity(contains_face=True).shape[0]
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid = _curvilinear_cube(n_lons, n_lats, lon_bounds, lat_bounds)

    h = 2
    t = 3
    height = DimCoord(np.arange(h), standard_name="height")
    time = DimCoord(np.arange(t), standard_name="time")

    src_data = np.empty([t, n_lats, n_lons, h])
    src_data[:] = np.arange(t * h).reshape([t, h])[:, np.newaxis, np.newaxis, :]
    cube = Cube(src_data)
    cube.add_aux_coord(grid.coord("latitude"), [1, 2])
    cube.add_aux_coord(grid.coord("longitude"), [1, 2])
    cube.add_dim_coord(time, 0)
    cube.add_dim_coord(height, 3)

    regridder = GridToMeshESMFRegridder(grid, tgt)
    cube_lazy = cube.copy(da.array(cube.data))
    result = regridder(cube)
    result_lazy = regridder(cube_lazy)

    # Lenient check for data.
    expected_data = np.empty([t, mesh_length, h])
    expected_data[:] = np.arange(t * h).reshape(t, h)[:, np.newaxis, :]
    assert np.allclose(expected_data, result.data)

    mesh_coord_x, mesh_coord_y = mesh.to_MeshCoords("face")
    expected_cube = Cube(expected_data)
    expected_cube.add_dim_coord(time, 0)
    expected_cube.add_aux_coord(mesh_coord_x, 1)
    expected_cube.add_aux_coord(mesh_coord_y, 1)
    expected_cube.add_dim_coord(height, 2)

    # Check metadata and scalar coords.
    result.data = expected_data
    assert expected_cube == result
    assert result_lazy == result


def test_mesh_target():
    """
    Basic test for :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Tests with a mesh as the target.
    """
    n_tgt_lons = 5
    n_tgt_lats = 4
    tgt = _gridlike_mesh(n_tgt_lons, n_tgt_lats)

    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    src = _add_metadata(src)
    src.data[:] = 1  # Ensure all data in the source is one.
    regridder = GridToMeshESMFRegridder(src, tgt, tgt_location="face")
    result = regridder(src)

    expected_data = np.ones([n_tgt_lats * n_tgt_lons])
    expected_cube = _add_metadata(_gridlike_mesh_cube(n_tgt_lons, n_tgt_lats))

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)

    # Check metadata and scalar coords.
    expected_cube.data = result.data
    assert expected_cube == result


@pytest.mark.parametrize(
    "resolution", (None, 2), ids=("no resolution", "with resolution")
)
def test_masks(resolution):
    """
    Test initialisation of :class:`esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`.

    Checks that the `use_src_mask` and `use_tgt_mask` keywords work properly.
    """
    if resolution is None:
        src = _curvilinear_cube(7, 6, [-180, 180], [-90, 90])
    else:
        # The resolution keyword is only valid for rectilinear grids.
        src = _grid_cube(7, 6, [-180, 180], [-90, 90])
    tgt = _gridlike_mesh_cube(6, 7)

    # Make src discontiguous at (0, 0)
    src_mask = np.zeros([6, 7], dtype=bool)
    src_mask[0, 0] = True
    src.data = np.ma.array(src.data, mask=src_mask)
    src_discontiguous = src.copy()
    if resolution is None:
        src_discontiguous.coord("latitude").bounds[0, 0] = 0
        src_discontiguous.coord("longitude").bounds[0, 0] = 0

    tgt_mask = np.zeros([7 * 6], dtype=bool)
    tgt_mask[0] = True
    tgt.data = np.ma.array(tgt.data, mask=tgt_mask)

    rg_src_masked = GridToMeshESMFRegridder(
        src_discontiguous, tgt, use_src_mask=True, src_resolution=resolution
    )
    rg_tgt_masked = GridToMeshESMFRegridder(
        src, tgt, use_tgt_mask=True, src_resolution=resolution
    )
    rg_unmasked = GridToMeshESMFRegridder(src, tgt, src_resolution=resolution)

    weights_src_masked = rg_src_masked.regridder.weight_matrix
    weights_tgt_masked = rg_tgt_masked.regridder.weight_matrix
    weights_unmasked = rg_unmasked.regridder.weight_matrix

    # Check there are no weights associated with the masked point.
    assert weights_src_masked[:, 0].nnz == 0
    assert weights_tgt_masked[0].nnz == 0

    # Check all other weights are correct.
    assert np.allclose(
        weights_src_masked[:, 1:].todense(), weights_unmasked[:, 1:].todense()
    )
    assert np.allclose(weights_tgt_masked[1:].todense(), weights_unmasked[1:].todense())



"""Unit tests for :class:`esmf_regrid.schemes._ESMFRegridder`."""

import pytest

from esmf_regrid.schemes import _ESMFRegridder
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _grid_cube,
)


def test_invalid_method():
    """
    Test initialisation of :class:`esmf_regrid.schemes._ESMFRegridder`.

    Checks that an error is raised when the method is invalid.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    with pytest.raises(ValueError):
        _ = _ESMFRegridder(src, tgt, method="other")
